import os
import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import cumtrapz, quad
from scipy.interpolate import interp1d, CubicSpline
from scipy.optimize import minimize_scalar

from .tov_solver import IntegrateTOV
from ...core import utils

C_SI = utils.speed_of_light  # m/s
C_CGS = C_SI * 100.
G_SI = utils.gravitational_constant  # m^3 kg^-1 s^-2
MSUN_SI = utils.solar_mass  # Kg

# Stores conversions from geometerized to cgs or si unit systems
conversion_dict = {'pressure': {'cgs': C_SI ** 4. / G_SI * 10., 'si': C_SI ** 4. / G_SI, 'geom': 1.},
                   'energy_density': {'cgs': C_SI ** 4. / G_SI * 10., 'si': C_SI ** 4. / G_SI, 'geom': 1.},
                   'density': {'cgs': C_SI ** 2. / G_SI / 1000., 'si': C_SI ** 2. / G_SI, 'geom': 1.},
                   'pseudo_enthalpy': {'dimensionless': 1.},
                   'mass': {'g': C_SI ** 2. / G_SI * 1000, 'kg': C_SI ** 2. / G_SI, 'geom': 1.,
                            'm_sol': C_SI ** 2. / G_SI / MSUN_SI},
                   'radius': {'cm': 100., 'mass': 1., 'km': .001},
                   'tidal_deformability': {'geom': 1.}}


# construct dictionary of pre-shipped EOS pressure denstity table
path_to_eos_tables = os.path.join(os.path.dirname(__file__), 'eos_tables')
list_of_eos_tables = os.listdir(path_to_eos_tables)
valid_eos_files = [i for i in list_of_eos_tables if 'LAL' in i]
valid_eos_file_paths = [os.path.join(path_to_eos_tables, filename) for filename in valid_eos_files]
valid_eos_names = [i.split('_', maxsplit=1)[-1].strip('.dat') for i in valid_eos_files]
valid_eos_dict = dict(zip(valid_eos_names, valid_eos_file_paths))


class TabularEOS(object):
    """
    Given a valid eos input format, such as 2-D array, an ascii file, or a string, parse, and interpolate

    Parameters
    ==========
    eos: (numpy.ndarray, str, ASCII TABLE)
        if `numpy.ndarray` then user supplied pressure-density 2D numpy array.
        if `str` then given a valid eos name, relevant preshipped ASCII table will be loaded
        if ASCII TABLE then given viable file extensions, which include .txt,.dat, etc (np.loadtxt used),
        read in pressure density from file.
    sampling_flag: bool
        Do you plan on sampling the parameterized EOS? Highly recommended. Defaults to False.
    warning_flag: bool
        Keeps track of status of various physical checks on EoS.

    Attributes
    ==========
    msg: str
        Human readable string describing the exception.
    code: int
        Exception error code.
    """

    def __init__(self, eos, sampling_flag=False, warning_flag=False):

        self.sampling_flag = sampling_flag
        self.warning_flag = warning_flag

        if type(eos) == str:
            if eos in valid_eos_dict.keys():
                table = np.loadtxt(valid_eos_dict[eos])
            else:
                table = np.loadtxt(eos)
        elif type(eos) == np.ndarray:
            table = eos
        else:
            raise ValueError("eos provided is invalid type please supply a str name, str path to ASCII file, "
                             "or a numpy array")

        table = self.__remove_leading_zero(table)

        # all have units of m^-2
        self.pressure = table[:, 0]
        self.energy_density = table[:, 1]

        self.minimum_pressure = min(self.pressure)
        self.minimum_energy_density = min(self.energy_density)
        if (not self.check_monotonicity() and self.sampling_flag) or self.warning_flag:
            self.warning_flag = True
        else:
            integrand = self.pressure / (self.energy_density + self.pressure)
            self.pseudo_enthalpy = cumtrapz(integrand, np.log(self.pressure), initial=0) + integrand[0]

            self.interp_energy_density_from_pressure = CubicSpline(np.log10(self.pressure),
                                                                   np.log10(self.energy_density),
                                                                   )

            self.interp_energy_density_from_pseudo_enthalpy = CubicSpline(np.log10(self.pseudo_enthalpy),
                                                                          np.log10(self.energy_density))

            self.interp_pressure_from_pseudo_enthalpy = CubicSpline(np.log10(self.pseudo_enthalpy),
                                                                    np.log10(self.pressure))

            self.interp_pseudo_enthalpy_from_energy_density = CubicSpline(np.log10(self.energy_density),
                                                                          np.log10(self.pseudo_enthalpy))

            self.__construct_all_tables()

            self.minimum_pseudo_enthalpy = min(self.pseudo_enthalpy)
            if not self.check_causality() and self.sampling_flag:
                self.warning_flag = True

    def __remove_leading_zero(self, table):
        """
        For interpolation of lalsimulation tables;
        loglog interpolation breaks if the first entries are 0s
        """

        if table[0, 0] == 0. or table[0, 1] == 0.:
            return table[1:, :]

        else:
            return table

    def energy_from_pressure(self, pressure, interp_type='CubicSpline'):
        """
        Find value of energy_from_pressure
        as in lalsimulation, return e = K * p**(3./5.) below min pressure

        Parameters
        ==========
        pressure: float
            pressure in geometerized units.
        interp_type: str
            String specifying which interpolation type to use.
            Currently implemented: 'CubicSpline', 'linear'.
        energy_density: float
            energy-density in geometerized units.
        """
        pressure = np.atleast_1d(pressure)
        energy_returned = np.zeros(pressure.size)
        indices_less_than_min = np.nonzero(pressure < self.minimum_pressure)
        indices_greater_than_min = np.nonzero(pressure >= self.minimum_pressure)

        # We do this special for less than min pressure
        energy_returned[indices_less_than_min] = 10 ** (np.log10(self.energy_density[0]) +
                                                        (3. / 5.) * (np.log10(pressure[indices_less_than_min]) -
                                                                     np.log10(self.pressure[0])))

        if interp_type == 'CubicSpline':
            energy_returned[indices_greater_than_min] = (
                10. ** self.interp_energy_density_from_pressure(np.log10(pressure[indices_greater_than_min])))
        elif interp_type == 'linear':
            energy_returned[indices_greater_than_min] = np.interp(pressure[indices_greater_than_min],
                                                                  self.pressure,
                                                                  self.energy_density)
        else:
            raise ValueError('Interpolation scheme must be linear or CubicSpline')

        if energy_returned.size == 1:
            return energy_returned[0]
        else:
            return energy_returned

    def pressure_from_pseudo_enthalpy(self, pseudo_enthalpy, interp_type='CubicSpline'):
        """
        Find p(h)
        as in lalsimulation, return p = K * h**(5./2.) below min enthalpy

        :param pseudo_enthalpy (`float`): Dimensionless pseudo-enthalpy.
        :interp_type (`str`): String specifying interpolation type.
                              Current implementations are 'CubicSpline', 'linear'.

        :return pressure (`float`): pressure in geometerized units.
        """
        pseudo_enthalpy = np.atleast_1d(pseudo_enthalpy)
        pressure_returned = np.zeros(pseudo_enthalpy.size)
        indices_less_than_min = np.nonzero(pseudo_enthalpy < self.minimum_pseudo_enthalpy)
        indices_greater_than_min = np.nonzero(pseudo_enthalpy >= self.minimum_pseudo_enthalpy)

        pressure_returned[indices_less_than_min] = 10. ** (np.log10(self.pressure[0]) +
                                                           2.5 * (np.log10(pseudo_enthalpy[indices_less_than_min]) -
                                                                  np.log10(self.pseudo_enthalpy[0])))

        if interp_type == 'CubicSpline':
            pressure_returned[indices_greater_than_min] = (
                10. ** self.interp_pressure_from_pseudo_enthalpy(np.log10(pseudo_enthalpy[indices_greater_than_min])))
        elif interp_type == 'linear':
            pressure_returned[indices_greater_than_min] = np.interp(pseudo_enthalpy[indices_greater_than_min],
                                                                    self.pseudo_enthalpy,
                                                                    self.pressure)
        else:
            raise ValueError('Interpolation scheme must be linear or CubicSpline')

        if pressure_returned.size == 1:
            return pressure_returned[0]
        else:
            return pressure_returned

    def energy_density_from_pseudo_enthalpy(self, pseudo_enthalpy, interp_type='CubicSpline'):
        """
        Find energy_density_from_pseudo_enthalpy(pseudo_enthalpy)
        as in lalsimulation, return e = K * h**(3./2.) below min enthalpy

        :param pseudo_enthalpy (`float`): Dimensionless pseudo-enthalpy.
        :param interp_type (`str`): String specifying interpolation type.
                                    Current implementations are 'CubicSpline', 'linear'.

        :return energy_density (`float`): energy-density in geometerized units.
        """
        pseudo_enthalpy = np.atleast_1d(pseudo_enthalpy)
        energy_returned = np.zeros(pseudo_enthalpy.size)
        indices_less_than_min = np.nonzero(pseudo_enthalpy < self.minimum_pseudo_enthalpy)
        indices_greater_than_min = np.nonzero(pseudo_enthalpy >= self.minimum_pseudo_enthalpy)

        energy_returned[indices_less_than_min] = 10 ** (np.log10(self.energy_density[0]) +
                                                        1.5 * (np.log10(pseudo_enthalpy[indices_less_than_min]) -
                                                               np.log10(self.pseudo_enthalpy[0])))
        if interp_type == 'CubicSpline':
            x = np.log10(pseudo_enthalpy[indices_greater_than_min])
            energy_returned[indices_greater_than_min] = 10 ** self.interp_energy_density_from_pseudo_enthalpy(x)
        elif interp_type == 'linear':
            energy_returned[indices_greater_than_min] = np.interp(pseudo_enthalpy[indices_greater_than_min],
                                                                  self.pseudo_enthalpy,
                                                                  self.energy_density)
        else:
            raise ValueError('Interpolation scheme must be linear or CubicSpline')

        if energy_returned.size == 1:
            return energy_returned[0]
        else:
            return energy_returned

    def pseudo_enthalpy_from_energy_density(self, energy_density, interp_type='CubicSpline'):
        """
        Find h(epsilon)
        as in lalsimulation, return h = K * e**(2./3.) below min enthalpy

        :param energy_density (`float`): energy-density in geometerized units.
        :param interp_type (`str`): String specifying interpolation type.
                                    Current implementations are 'CubicSpline', 'linear'.

        :return pseudo_enthalpy (`float`): Dimensionless pseudo-enthalpy.
        """
        energy_density = np.atleast_1d(energy_density)
        pseudo_enthalpy_returned = np.zeros(energy_density.size)
        indices_less_than_min = np.nonzero(energy_density < self.minimum_energy_density)
        indices_greater_than_min = np.nonzero(energy_density >= self.minimum_energy_density)

        pseudo_enthalpy_returned[indices_less_than_min] = 10 ** (np.log10(self.pseudo_enthalpy[0]) + (2. / 3.) *
                                                                 (np.log10(energy_density[indices_less_than_min]) -
                                                                  np.log10(self.energy_density[0])))

        if interp_type == 'CubicSpline':
            x = np.log10(energy_density[indices_greater_than_min])
            pseudo_enthalpy_returned[indices_greater_than_min] = 10**self.interp_pseudo_enthalpy_from_energy_density(x)
        elif interp_type == 'linear':
            pseudo_enthalpy_returned[indices_greater_than_min] = np.interp(energy_density[indices_greater_than_min],
                                                                           self.energy_density,
                                                                           self.pseudo_enthalpy)
        else:
            raise ValueError('Interpolation scheme must be linear or CubicSpline')

        if pseudo_enthalpy_returned.size == 1:
            return pseudo_enthalpy_returned[0]
        else:
            return pseudo_enthalpy_returned

    def dedh(self, pseudo_enthalpy, rel_dh=1e-5, interp_type='CubicSpline'):
        """
        Value of [depsilon/dh](p)

        :param pseudo_enthalpy (`float`): Dimensionless pseudo-enthalpy.
        :param interp_type (`str`): String specifying interpolation type.
                                    Current implementations are 'CubicSpline', 'linear'.
        :param rel_dh (`float`): Relative step size in pseudo-enthalpy space.

        :return dedh (`float`): Derivative of energy-density with respect to pseudo-enthalpy
                                evaluated at `pseudo_enthalpy` in geometerized units.
        """

        # step size=fraction of value
        dh = pseudo_enthalpy * rel_dh

        eps_upper = self.energy_density_from_pseudo_enthalpy(pseudo_enthalpy + dh, interp_type=interp_type)
        eps_lower = self.energy_density_from_pseudo_enthalpy(pseudo_enthalpy - dh, interp_type=interp_type)

        return (eps_upper - eps_lower) / (2. * dh)

    def dedp(self, pressure, rel_dp=1e-5, interp_type='CubicSpline'):
        """
        Find value of [depsilon/dp](p)

        :param pressure (`float`): pressure in geometerized units.
        :param rel_dp (`float`): Relative step size in pressure space.
        :param interp_type (`float`): String specifying interpolation type.
                                      Current implementations are 'CubicSpline', 'linear'.

        :return dedp (`float`): Derivative of energy-density with respect to pressure
                                evaluated at `pressure`.
        """

        # step size=fraction of value
        dp = pressure * rel_dp

        eps_upper = self.energy_from_pressure(pressure + dp, interp_type=interp_type)
        eps_lower = self.energy_from_pressure(pressure - dp, interp_type=interp_type)

        return (eps_upper - eps_lower) / (2. * dp)

    def velocity_from_pseudo_enthalpy(self, pseudo_enthalpy, interp_type='CubicSpline'):
        """
        Returns the speed of sound in geometerized units in the
        neutron star at the specified pressure.

        Assumes the equation
        vs = c (de/dp)^{-1/2}

        :param pseudo_enthalpy (`float`): Dimensionless pseudo-enthalpy.
        :param interp_type (`str`): String specifying interpolation type.
                                    Current implementations are 'CubicSpline', 'linear'.

        :return v_s (`float`): Speed of sound at `pseudo-enthalpy` in geometerized units.
        """
        pressure = self.pressure_from_pseudo_enthalpy(pseudo_enthalpy, interp_type=interp_type)
        return self.dedp(pressure, interp_type=interp_type) ** -0.5

    def check_causality(self):
        """
        Checks to see if the equation of state is causal i.e. the speed
        of sound in the star is less than the speed of light.
        Returns True if causal, False if not.
        """
        pmax = self.pressure[-1]
        emax = self.energy_from_pressure(pmax)
        hmax = self.pseudo_enthalpy_from_energy_density(emax)
        vsmax = self.velocity_from_pseudo_enthalpy(hmax)
        if vsmax < 1.1:
            return True
        else:
            return False

    def check_monotonicity(self):
        """
        Checks to see if the equation of state is monotonically increasing
        in energy density-pressure space. Returns True if monotonic, False if not.
        """
        e1 = self.energy_density[1:]
        e2 = self.energy_density[:-1]
        ediff = e1 - e2
        e_negatives = len(np.where(ediff < 0))

        p1 = self.pressure[1:]
        p2 = self.pressure[:-1]
        pdiff = p1 - p2
        p_negatives = len(np.where(pdiff < 0))
        if e_negatives > 1 or p_negatives > 1:
            return False
        else:
            return True

    def __get_plot_range(self, data):
        """
        Determines default plot range based on data provided.
        """
        low = np.amin(data)
        high = np.amax(data)
        dx = 0.05 * (high - low)

        xmin = low - dx
        xmax = high + dx
        xlim = [xmin, xmax]

        return xlim

    def __construct_all_tables(self):
        """Pressure and epsilon already tabular, now create array of enthalpies"""
        edat = self.energy_density
        hdat = [self.pseudo_enthalpy_from_energy_density(e) for e in edat]
        self.pseudo_enthalpy = np.array(hdat)

    def plot(self, rep, xlim=None, ylim=None, units=None):
        """
        Given a representation in the form 'energy_density-pressure', plot the EoS in that space.

        Parameters
        ==========
        rep: str
            Representation to plot. For example, plotting in energy_density-pressure space,
            specify 'energy_density-pressure'
        xlim: list
            Plotting bounds for x-axis in the form [low, high].
            Defaults to 'None' which will plot from 10% below min x value to 10% above max x value
        ylim: list
            Plotting bounds for y-axis in the form [low, high].
            Defaults to 'None' which will plot from 10% below min y value to 10% above max y value
        units: str
            Specifies unit system to plot. Currently can plot in CGS:'cgs', SI:'si', or geometerized:'geom'

        Returns
        =======
        fig: matplotlib.figure.Figure
            EOS plot.
        """

        # Set data based on specified representation
        varnames = rep.split('-')

        assert varnames[0] != varnames[
            1], 'Cannot plot the same variable against itself. Please choose another representation'

        # Correspondence of rep parameter, data, and latex symbol
        # rep_dict = {'energy_density': [self.epsilon, r'$\epsilon$'],
        # 'pressure': [self.p, r'$p$'], 'pseudo_enthalpy': [pseudo_enthalpy, r'$h$']}

        # FIXME: The second element in these arrays should be tex labels, but tex's not working rn
        rep_dict = {'energy_density': [self.energy_density, 'energy_density'],
                    'pressure': [self.pressure, 'pressure'],
                    'pseudo_enthalpy': [self.pseudo_enthalpy, 'pseudo_enthalpy']}

        xname = varnames[1]
        yname = varnames[0]

        # Set units
        eos_default_units = {'pressure': 'cgs', 'energy_density': 'cgs', 'density': 'cgs',
                             'pseudo_enthalpy': 'dimensionless'}
        if units is None:
            units = [eos_default_units[yname], eos_default_units[xname]]          # Default unit system is cgs
        elif isinstance(units, str):
            units = [units, units]          # If only one unit system given, use for both

        xunits = units[1]
        yunits = units[0]

        # Ensure valid units
        if xunits not in list(conversion_dict[xname].keys()) or yunits not in list(conversion_dict[yname].keys()):
            s = '''
                Invalid unit system. Valid variable-unit pairs are:
                p: {p_units}
                e: {e_units}
                rho: {rho_units}
                h: {h_units}.
                '''.format(p_units=list(conversion_dict['pressure'].keys()),
                           e_units=list(conversion_dict['energy_density'].keys()),
                           rho_units=list(conversion_dict['density'].keys()),
                           h_units=list(conversion_dict['pseudo_enthalpy'].keys()))
            raise ValueError(s)

        xdat = rep_dict[xname][0] * conversion_dict[xname][xunits]
        ydat = rep_dict[yname][0] * conversion_dict[yname][yunits]

        xlabel = rep_dict[varnames[1]][1].replace('_', ' ')
        ylabel = rep_dict[varnames[0]][1].replace('_', ' ') + '(' + xlabel + ')'

        # Determine plot ranges. Currently shows 10% wider than actual data range.
        if xlim is None:
            xlim = self.__get_plot_range(xdat)

        if ylim is None:
            ylim = self.__get_plot_range(ydat)

        fig, ax = plt.subplots()

        ax.loglog(xdat, ydat)
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        return fig


def spectral_adiabatic_index(gammas, x):
    arg = 0
    for i in range(len(gammas)):
        arg += gammas[i] * x ** i

    return np.exp(arg)


class SpectralDecompositionEOS(TabularEOS):
    """
    Parameterized EOS using a spectral
    decomposition per Lindblom
    arXiv: 1009.0738v2. Inherits from TabularEOS.

    Parameters
    ==========
    gammas: list
        List of adiabatic expansion parameters used
        to construct the equation of state in various
        spaces.
    p0: float
        The starting point in pressure of the high-density EoS. This is stitched to
        the low-density portion of the SLY EoS model. The default value chosen is set to
        a sufficiently low pressure so that the high-density EoS will never be
        overconstrained.
    e0/c**2: float
        The starting point in energy-density of the high-density EoS. This is stitched to
        the low-density portion of the SLY EoS model. The default value chosen is set to
        a sufficiently low energy density so that the high-density EoS will never be
        overconstrained.
    xmax: float
        highest dimensionless pressure value in EoS
    npts: float (optional)
        number of points in pressure-energy density data.
    """

    def __init__(self, gammas, p0=3.01e33, e0=2.03e14, xmax=None, npts=100, sampling_flag=False, warning_flag=False):
        self.warning_flag = warning_flag
        self.gammas = gammas
        self.p0 = p0
        self.e0 = e0
        self.xmax = xmax
        self.npts = npts
        if self.xmax is None:
            self.xmax = self.__determine_xmax()
        self.sampling_flag = sampling_flag
        self.__construct_a_of_x_table()

        # Construct pressure-energy density table and
        # set up interpolating functions.

        if self.warning_flag and self.sampling_flag:
            # If sampling prior is enabled and adiabatic check
            # has failed, empty the array values
            self.e_pdat = np.zeros((2, 2))
        else:
            self.e_pdat = self.__construct_e_of_p_table()

        super().__init__(self.e_pdat, sampling_flag=self.sampling_flag, warning_flag=self.warning_flag)

    def __determine_xmax(self, a_max=6.0):
        highest_order_gamma = np.abs(self.gammas[-1])[0]
        expansion_order = float(len(self.gammas) - 1)

        xmax = (np.log(a_max) / highest_order_gamma) ** (1. / expansion_order)
        return xmax

    def __mu_integrand(self, x):

        return 1. / spectral_adiabatic_index(self.gammas, x)

    def mu(self, x):

        return np.exp(-quad(self.__mu_integrand, 0, x)[0])

    def __eps_integrand(self, x):

        return np.exp(x) * self.mu(x) / spectral_adiabatic_index(self.gammas, x)

    def energy_density(self, x, eps0):

        quad_result, quad_err = quad(self.__eps_integrand, 0, x)
        eps_of_x = (eps0 * C_CGS ** 2.) / self.mu(x) + self.p0 / self.mu(x) * quad_result
        return eps_of_x

    def __construct_a_of_x_table(self):

        xdat = np.linspace(0, self.xmax, num=self.npts)

        # Generate adiabatic index points until a point is out of prior range
        if self.sampling_flag:
            adat = np.empty(self.npts)
            for i in range(self.npts):
                if 0.6 < spectral_adiabatic_index(self.gammas, xdat[i]) < 4.6:
                    adat[i] = spectral_adiabatic_index(self.gammas, xdat[i])
                else:
                    break

            # Truncate arrays to last point within range
            adat = adat[:i]
            xmax_new = xdat[i - 1]

            # If EOS is too short, set prior to 0, else resample the function and set new xmax
            if xmax_new < 4. or i == 0:
                self.warning_flag = True
            else:
                xdat = np.linspace(0, xmax_new, num=self.npts)
                adat = spectral_adiabatic_index(self.gammas, xdat)
                self.xmax = xmax_new
        else:
            adat = spectral_adiabatic_index(self.gammas, xdat)

        self.adat = adat

    def __construct_e_of_p_table(self):

        """
        Creates p, epsilon table for a given set of spectral parameters
        """

        # make p range
        # to match lalsimulation tables: array = [pressure, density]
        x_range = np.linspace(0, self.xmax, self.npts)
        p_range = self.p0 * np.exp(x_range)

        eos_vals = np.zeros((self.npts, 2))
        eos_vals[:, 0] = p_range

        for i in range(0, len(x_range)):
            eos_vals[i, 1] = self.energy_density(x_range[i], self.e0)

        # convert eos to geometrized units in *m^-2*
        # IMPORTANT
        eos_vals = eos_vals * 0.1 * G_SI / C_SI ** 4

        # doing as those before me have done and using SLY4 as low density region
        # SLY4 in geometrized units
        low_density_path = os.path.join(os.path.dirname(__file__), 'eos_tables', 'LALSimNeutronStarEOS_SLY4.dat')
        low_density = np.loadtxt(low_density_path)

        cutoff = eos_vals[0, :]

        # Then find overlap point
        break_pt = len(low_density)
        for i in range(1, len(low_density)):
            if low_density[-i, 0] < cutoff[0] and low_density[-i, 1] < cutoff[1]:
                break_pt = len(low_density) - i + 1
                break

        # stack EOS arrays
        eos_vals = np.vstack((low_density[0:break_pt, :], eos_vals))

        return eos_vals


class EOSFamily(object):
    """
    Create a EOS family and get mass-radius information

    Parameters
    ==========
    eos: object
        Supply a `TabularEOS` class (or subclass)
    npts: float
        Number of points to calculate for mass-radius relation. Default is 500.

    Notes
    =====
    The mass-radius and mass-k2 data should be
    populated here via the TOV solver upon object construction.
    """
    def __init__(self, eos, npts=500):
        self.eos = eos

        # FIXME: starting_energy_density is set somewhat arbitrarily
        starting_energy_density = 1.6e-10
        ending_energy_density = max(self.eos.energy_density)
        log_starting_energy_density = np.log(starting_energy_density)
        log_ending_energy_density = np.log(ending_energy_density)
        log_energy_density_grid = np.linspace(log_starting_energy_density,
                                              log_ending_energy_density,
                                              num=npts)
        energy_density_grid = np.exp(log_energy_density_grid)

        # Generate m, r, and k2 lists
        mass = []
        radius = []
        k2love_number = []
        for i in range(len(energy_density_grid)):
            tov_solver = IntegrateTOV(self.eos, energy_density_grid[i])
            m, r, k2 = tov_solver.integrate_TOV()
            mass.append(m)
            radius.append(r)
            k2love_number.append(k2)

            # Check if maximum mass has been found
            if i > 0 and mass[i] <= mass[i - 1]:
                break

        # If we're not at the end of the array, determine actual maximum mass. Else, assume
        # last point is the maximum mass and proceed.
        if i < (npts - 1):
            # Now replace ith point with interpolated maximum mass
            # This is done by interpolating the last three points and then
            # minimizing the negative of the interpolated function
            x = [energy_density_grid[i - 2], energy_density_grid[i - 1], energy_density_grid[i]]
            y = [mass[i - 2], mass[i - 1], mass[i]]

            f = interp1d(x, y, kind='quadratic', bounds_error=False, fill_value='extrapolate')

            res = minimize_scalar(lambda x: -f(x))

            # Integrate max energy density to get maximum mass
            tov_solver = IntegrateTOV(self.eos, res.x)
            mfin, rfin, k2fin = tov_solver.integrate_TOV()

            mass[-1] = mfin
            radius[-1] = rfin
            k2love_number[-1] = k2fin

        # Currently, everything is in geometerized units.
        # The mass variables have dimensions of length, k2 is dimensionless
        # and radii have dimensions of length. Calculate dimensionless lambda
        # with these quantities, then convert to SI.

        # Calculating dimensionless lambda values from k2, radii, and mass
        tidal_deformability = [2. / 3. * k2 * r ** 5. / m ** 5. for k2, r, m in
                               zip(k2love_number, radius, mass)]

        # As a last resort, if highest mass is still smaller than second
        # to last point, remove the last point from each array
        if mass[-1] < mass[-2]:
            mass = mass[:-1]
            radius = radius[:-1]
            k2love_number = k2love_number[:-1]
            tidal_deformability = tidal_deformability[:-1]

        self.mass = np.array(mass)
        self.radius = np.array(radius)
        self.k2love_number = np.array(k2love_number)
        self.tidal_deformability = np.array(tidal_deformability)
        self.maximum_mass = mass[-1] * conversion_dict['mass']['m_sol']

    def radius_from_mass(self, m):
        """
        :param m: mass of neutron star in solar masses
        :return: radius of neutron star in meters
        """
        f = CubicSpline(self.mass, self.radius, bc_type='natural', extrapolate=True)

        mass_converted_to_geom = m * MSUN_SI * G_SI / C_SI ** 2.
        return f(mass_converted_to_geom)

    def k2_from_mass(self, m):
        """
        :param m: mass of neutron star in solar masses.
        :return: dimensionless second tidal love number.
        """
        f = CubicSpline(self.mass, self.k2love_number, bc_type='natural', extrapolate=True)

        m_geom = m * MSUN_SI * G_SI / C_SI ** 2.
        return f(m_geom)

    def lambda_from_mass(self, m):
        """
        Convert from equation of state model parameters to
        component tidal parameters.

        :param m: Mass of neutron star in solar masses.
        :return: Tidal parameter of neutron star of mass m.
        """

        # Get lambda from mass and equation of state

        r = self.radius_from_mass(m)
        k = self.k2_from_mass(m)
        m_geom = m * MSUN_SI * G_SI / C_SI ** 2.
        c = m_geom / r
        lmbda = (2. / 3.) * k / c ** 5.

        return lmbda

    def __get_plot_range(self, data):
        low = np.amin(data)
        high = np.amax(data)
        dx = 0.05 * (high - low)

        xmin = low - dx
        xmax = high + dx
        xlim = [xmin, xmax]

        return xlim

    def plot(self, rep, xlim=None, ylim=None, units=None):
        """
        Given a representation in the form 'm-r', plot the family in that space.

        Parameters
        ==========
        rep: str
            Representation to plot. For example, plotting in mass-radius space, specify 'm-r'
        xlim: list
            Plotting bounds for x-axis in the form [low, high].
            Defaults to 'None' which will plot from 10% below min x value to 10% above max x value
        ylim: list
            Plotting bounds for y-axis in the form [low, high].
            Defaults to 'None' which will plot from 10% below min y value to 10% above max y value
        units: str
            Specifies unit system to plot. Currently can plot in CGS:'cgs', SI:'si', or geometerized:'geom'

        Returns
        =======
        fig: matplotlib.figure.Figure
            EOS Family plot.
        """

        # Set data based on specified representation
        varnames = rep.split('-')

        assert varnames[0] != varnames[
            1], 'Cannot plot the same variable against itself. Please choose another representation'

        # Correspondence of rep parameter, data, and latex symbol
        rep_dict = {'mass': [self.mass, r'$M$'], 'radius': [self.radius, r'$R$'], 'k2': [self.k2love_number, r'$k_2$'],
                    'tidal_deformability': [self.tidal_deformability, r'$l$']}

        xname = varnames[1]
        yname = varnames[0]

        # Set units
        fam_default_units = {'mass': 'm_sol', 'radius': 'km', 'tidal_deformability': 'geom'}
        if units is None:
            units = [fam_default_units[yname], fam_default_units[xname]]  # Default unit system is cgs
        elif isinstance(units, str):
            units = [units, units]  # If only one unit system given, use for both

        xunits = units[1]
        yunits = units[0]

        # Ensure valid units
        if xunits not in list(conversion_dict[xname].keys()) or yunits not in list(conversion_dict[yname].keys()):
            s = '''
                        Invalid unit system. Valid variable-unit pairs are:
                        m: {m_units}
                        r: {r_units}
                        l: {l_units}.
                        '''.format(m_units=list(conversion_dict['mass'].keys()),
                                   r_units=list(conversion_dict['radius'].keys()),
                                   l_units=list(conversion_dict['tidal_deformability'].keys()))
            raise ValueError(s)

        xdat = rep_dict[varnames[1]][0] * conversion_dict[xname][xunits]
        ydat = rep_dict[varnames[0]][0] * conversion_dict[yname][yunits]

        xlabel = rep_dict[varnames[1]][1].replace('_', ' ')
        ylabel = rep_dict[varnames[0]][1].replace('_', ' ') + '(' + xlabel + ')'

        # Determine plot ranges. Currently shows 10% wider than actual data range.
        if xlim is None:
            xlim = self.__get_plot_range(xdat)

        if ylim is None:
            ylim = self.__get_plot_range(ydat)

        # Plot the data with appropriate labels
        fig, ax = plt.subplots()

        ax.loglog(xdat, ydat)
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        return fig



import os
import unittest

import numpy as np
from past.builtins import execfile

import bilby


class Test(unittest.TestCase):
    outdir = "./outdir"
    dir_path = os.path.dirname(os.path.realpath(__file__))
    dir_path = os.path.abspath(os.path.join(dir_path, os.path.pardir))

    # Run a script to produce standard data
    msd = {}  # A dictionary of variables saved in make_standard_data.py
    execfile(dir_path + "/integration/make_standard_data.py", msd)
    """
    @classmethod
    def setUpClass(self):
        if os.path.isdir(self.outdir):
            try:
                shutil.rmtree(self.outdir)
            except OSError:
                logging.warning(
                    "{} not removed prior to tests".format(self.outdir))

    @classmethod
    def tearDownClass(self):
        if os.path.isdir(self.outdir):
            try:
                shutil.rmtree(self.outdir)
            except OSError:
                logging.warning(
                    "{} not removed prior to tests".format(self.outdir))
    """

    def test_make_standard_data(self):
        " Load in the saved standard data and compare with new data "

        # Load in the saved standard data
        frequencies_saved, hf_real_saved, hf_imag_saved = np.loadtxt(
            self.dir_path + "/integration/standard_data.txt"
        ).T
        hf_signal_and_noise_saved = hf_real_saved + 1j * hf_imag_saved

        self.assertTrue(np.array_equal(self.msd["frequencies"], frequencies_saved))
        self.assertAlmostEqual(
            all(self.msd["hf_signal_and_noise"] - hf_signal_and_noise_saved),
            0.00000000,
            5,
        )

    def test_recover_luminosity_distance(self):
        likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
            [self.msd["IFO"]], self.msd["waveform_generator"]
        )

        priors = {}
        for key in self.msd["simulation_parameters"]:
            priors[key] = self.msd["simulation_parameters"][key]

        dL = self.msd["simulation_parameters"]["luminosity_distance"]
        priors["luminosity_distance"] = bilby.core.prior.Uniform(
            name="luminosity_distance", minimum=dL - 10, maximum=dL + 10
        )

        result = bilby.core.sampler.run_sampler(
            likelihood, priors, sampler="dynesty", verbose=False, npoints=100
        )
        self.assertAlmostEqual(
            np.mean(result.posterior.luminosity_distance),
            dL,
            delta=3 * np.std(result.posterior.luminosity_distance),
        )


if __name__ == "__main__":
    unittest.main()



import os
import shutil
import unittest

import pandas as pd

import bilby


class TestCBCResult(unittest.TestCase):
    def setUp(self):
        bilby.utils.command_line_args.bilby_test_mode = False
        priors = bilby.gw.prior.BBHPriorDict()
        priors["geocent_time"] = 2
        injection_parameters = priors.sample()
        self.meta_data = dict(
            likelihood=dict(
                phase_marginalization=True,
                distance_marginalization=False,
                time_marginalization=True,
                frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
                waveform_arguments=dict(
                    reference_frequency=20.0, waveform_approximant="IMRPhenomPv2"
                ),
                interferometers=dict(
                    H1=dict(optimal_SNR=1, parameters=injection_parameters),
                    L1=dict(optimal_SNR=1, parameters=injection_parameters),
                ),
                sampling_frequency=4096,
                duration=4,
                start_time=0,
                waveform_generator_class=bilby.gw.waveform_generator.WaveformGenerator,
                parameter_conversion=bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters,
            )
        )
        self.result = bilby.gw.result.CBCResult(
            label="label",
            outdir="outdir",
            sampler="nestle",
            search_parameter_keys=list(priors.keys()),
            fixed_parameter_keys=list(),
            priors=priors,
            sampler_kwargs=dict(test="test", func=lambda x: x),
            injection_parameters=injection_parameters,
            meta_data=self.meta_data,
            posterior=pd.DataFrame(priors.sample(100)),
        )
        if not os.path.isdir(self.result.outdir):
            os.mkdir(self.result.outdir)

    def tearDown(self):
        bilby.utils.command_line_args.bilby_test_mode = True
        try:
            shutil.rmtree(self.result.outdir)
        except OSError:
            pass
        del self.result

    def test_phase_marginalization(self):
        self.assertEqual(
            self.result.phase_marginalization,
            self.meta_data["likelihood"]["phase_marginalization"],
        )

    def test_phase_marginalization_unset(self):
        self.result.meta_data["likelihood"].pop("phase_marginalization")
        with self.assertRaises(AttributeError):
            self.result.phase_marginalization

    def test_time_marginalization(self):
        self.assertEqual(
            self.result.time_marginalization,
            self.meta_data["likelihood"]["time_marginalization"],
        )

    def test_time_marginalization_unset(self):
        self.result.meta_data["likelihood"].pop("time_marginalization")
        with self.assertRaises(AttributeError):
            self.result.time_marginalization

    def test_distance_marginalization(self):
        self.assertEqual(
            self.result.distance_marginalization,
            self.meta_data["likelihood"]["distance_marginalization"],
        )

    def test_distance_marginalization_unset(self):
        self.result.meta_data["likelihood"].pop("distance_marginalization")
        with self.assertRaises(AttributeError):
            self.result.distance_marginalization

    def test_reference_frequency(self):
        self.assertEqual(
            self.result.reference_frequency,
            self.meta_data["likelihood"]["waveform_arguments"]["reference_frequency"],
        )

    def test_reference_frequency_unset(self):
        self.result.meta_data["likelihood"]["waveform_arguments"].pop(
            "reference_frequency"
        )
        with self.assertRaises(AttributeError):
            self.result.reference_frequency

    def test_sampling_frequency(self):
        self.assertEqual(
            self.result.sampling_frequency,
            self.meta_data["likelihood"]["sampling_frequency"],
        )

    def test_sampling_frequency_unset(self):
        self.result.meta_data["likelihood"].pop("sampling_frequency")
        with self.assertRaises(AttributeError):
            self.result.sampling_frequency

    def test_duration(self):
        self.assertEqual(self.result.duration, self.meta_data["likelihood"]["duration"])

    def test_duration_unset(self):
        self.result.meta_data["likelihood"].pop("duration")
        with self.assertRaises(AttributeError):
            self.result.duration

    def test_start_time(self):
        self.assertEqual(
            self.result.start_time, self.meta_data["likelihood"]["start_time"]
        )

    def test_start_time_unset(self):
        self.result.meta_data["likelihood"].pop("start_time")
        with self.assertRaises(AttributeError):
            self.result.start_time

    def test_waveform_approximant(self):
        self.assertEqual(
            self.result.waveform_approximant,
            self.meta_data["likelihood"]["waveform_arguments"]["waveform_approximant"],
        )

    def test_waveform_approximant_unset(self):
        self.result.meta_data["likelihood"]["waveform_arguments"].pop(
            "waveform_approximant"
        )
        with self.assertRaises(AttributeError):
            self.result.waveform_approximant

    def test_waveform_arguments(self):
        self.assertEqual(
            self.result.waveform_arguments,
            self.meta_data["likelihood"]["waveform_arguments"],
        )

    def test_frequency_domain_source_model(self):
        self.assertEqual(
            self.result.frequency_domain_source_model,
            self.meta_data["likelihood"]["frequency_domain_source_model"],
        )

    def test_frequency_domain_source_model_unset(self):
        self.result.meta_data["likelihood"].pop("frequency_domain_source_model")
        with self.assertRaises(AttributeError):
            self.result.frequency_domain_source_model

    def test_parameter_conversion(self):
        self.assertEqual(
            self.result.parameter_conversion,
            self.meta_data["likelihood"]["parameter_conversion"],
        )

    def test_parameter_conversion_unset(self):
        self.result.meta_data["likelihood"].pop("parameter_conversion")
        with self.assertRaises(AttributeError):
            self.result.parameter_conversion

    def test_waveform_generator_class(self):
        self.assertEqual(
            self.result.waveform_generator_class,
            self.meta_data["likelihood"]["waveform_generator_class"],
        )

    def test_waveform_generator_class_unset(self):
        self.result.meta_data["likelihood"].pop("waveform_generator_class")
        with self.assertRaises(AttributeError):
            self.result.waveform_generator_class

    def test_interferometer_names(self):
        self.assertEqual(
            self.result.interferometers,
            [name for name in self.meta_data["likelihood"]["interferometers"]],
        )

    def test_detector_injection_properties(self):
        self.assertEqual(
            self.result.detector_injection_properties("H1"),
            self.meta_data["likelihood"]["interferometers"]["H1"],
        )

    def test_detector_injection_properties_no_injection(self):
        self.assertEqual(
            self.result.detector_injection_properties("not_a_detector"), None
        )


if __name__ == "__main__":
    unittest.main()



import unittest

import mock

import bilby


class TestConditionalPrior(unittest.TestCase):
    def setUp(self):
        self.condition_func_call_counter = 0

        def condition_func(reference_parameters, test_variable_1, test_variable_2):
            self.condition_func_call_counter += 1
            return {key: value + 1 for key, value in reference_parameters.items()}

        self.condition_func = condition_func
        self.minimum = 0
        self.maximum = 5
        self.test_variable_1 = 0
        self.test_variable_2 = 1
        self.prior = bilby.core.prior.ConditionalBasePrior(
            condition_func=condition_func, minimum=self.minimum, maximum=self.maximum
        )

    def tearDown(self):
        del self.condition_func
        del self.condition_func_call_counter
        del self.minimum
        del self.maximum
        del self.test_variable_1
        del self.test_variable_2
        del self.prior

    def test_reference_params(self):
        self.assertDictEqual(
            dict(minimum=self.minimum, maximum=self.maximum),
            self.prior.reference_params,
        )

    def test_required_variables(self):
        self.assertListEqual(
            ["test_variable_1", "test_variable_2"],
            sorted(self.prior.required_variables),
        )

    def test_required_variables_no_condition_func(self):
        self.prior = bilby.core.prior.ConditionalBasePrior(
            condition_func=None, minimum=self.minimum, maximum=self.maximum
        )
        self.assertListEqual([], self.prior.required_variables)

    def test_get_instantiation_dict(self):
        expected = dict(
            minimum=0,
            maximum=5,
            name=None,
            latex_label=None,
            unit=None,
            boundary=None,
            condition_func=self.condition_func,
        )
        actual = self.prior.get_instantiation_dict()
        for key, value in expected.items():
            if key == "condition_func":
                continue
            self.assertEqual(value, actual[key])

    def test_update_conditions_correct_variables(self):
        self.prior.update_conditions(
            test_variable_1=self.test_variable_1, test_variable_2=self.test_variable_2
        )
        self.assertEqual(1, self.condition_func_call_counter)
        self.assertEqual(self.minimum + 1, self.prior.minimum)
        self.assertEqual(self.maximum + 1, self.prior.maximum)

    def test_update_conditions_no_variables(self):
        self.prior.update_conditions(
            test_variable_1=self.test_variable_1, test_variable_2=self.test_variable_2
        )
        self.prior.update_conditions()
        self.assertEqual(1, self.condition_func_call_counter)
        self.assertEqual(self.minimum + 1, self.prior.minimum)
        self.assertEqual(self.maximum + 1, self.prior.maximum)

    def test_update_conditions_illegal_variables(self):
        with self.assertRaises(bilby.core.prior.IllegalRequiredVariablesException):
            self.prior.update_conditions(test_parameter_1=self.test_variable_1)

    def test_sample_calls_update_conditions(self):
        with mock.patch.object(self.prior, "update_conditions") as m:
            self.prior.sample(
                1,
                test_parameter_1=self.test_variable_1,
                test_parameter_2=self.test_variable_2,
            )
            m.assert_called_with(
                test_parameter_1=self.test_variable_1,
                test_parameter_2=self.test_variable_2,
            )

    def test_rescale_calls_update_conditions(self):
        with mock.patch.object(self.prior, "update_conditions") as m:
            self.prior.rescale(
                1,
                test_parameter_1=self.test_variable_1,
                test_parameter_2=self.test_variable_2,
            )
            m.assert_called_with(
                test_parameter_1=self.test_variable_1,
                test_parameter_2=self.test_variable_2,
            )

    def test_prob_calls_update_conditions(self):
        with mock.patch.object(self.prior, "update_conditions") as m:
            self.prior.prob(
                1,
                test_parameter_1=self.test_variable_1,
                test_parameter_2=self.test_variable_2,
            )
            m.assert_called_with(
                test_parameter_1=self.test_variable_1,
                test_parameter_2=self.test_variable_2,
            )

    def test_rescale_ln_prob_update_conditions(self):
        with mock.patch.object(self.prior, "update_conditions") as m:
            self.prior.ln_prob(
                1,
                test_parameter_1=self.test_variable_1,
                test_parameter_2=self.test_variable_2,
            )
            calls = [
                mock.call(
                    test_parameter_1=self.test_variable_1,
                    test_parameter_2=self.test_variable_2,
                ),
                mock.call(),
            ]
            m.assert_has_calls(calls)

    def test_cdf_calls_update_conditions(self):
        self.prior = bilby.core.prior.ConditionalUniform(
            condition_func=self.condition_func, minimum=self.minimum, maximum=self.maximum
        )
        with mock.patch.object(self.prior, "update_conditions") as m:
            self.prior.cdf(
                1,
                test_parameter_1=self.test_variable_1,
                test_parameter_2=self.test_variable_2,
            )
            m.assert_called_with(
                test_parameter_1=self.test_variable_1,
                test_parameter_2=self.test_variable_2,
            )

    def test_reset_to_reference_parameters(self):
        self.prior.minimum = 10
        self.prior.maximum = 20
        self.prior.reset_to_reference_parameters()
        self.assertEqual(self.prior.reference_params["minimum"], self.prior.minimum)
        self.assertEqual(self.prior.reference_params["maximum"], self.prior.maximum)

    def test_cond_prior_instantiation_no_boundary_prior(self):
        prior = bilby.core.prior.ConditionalFermiDirac(
            condition_func=None, sigma=1, mu=1
        )
        self.assertIsNone(prior.boundary)


class TestConditionalPriorDict(unittest.TestCase):
    def setUp(self):
        def condition_func_1(reference_parameters, var_0):
            return reference_parameters

        def condition_func_2(reference_parameters, var_0, var_1):
            return reference_parameters

        def condition_func_3(reference_parameters, var_1, var_2):
            return reference_parameters

        self.minimum = 0
        self.maximum = 1
        self.prior_0 = bilby.core.prior.Uniform(
            minimum=self.minimum, maximum=self.maximum
        )
        self.prior_1 = bilby.core.prior.ConditionalUniform(
            condition_func=condition_func_1, minimum=self.minimum, maximum=self.maximum
        )
        self.prior_2 = bilby.core.prior.ConditionalUniform(
            condition_func=condition_func_2, minimum=self.minimum, maximum=self.maximum
        )
        self.prior_3 = bilby.core.prior.ConditionalUniform(
            condition_func=condition_func_3, minimum=self.minimum, maximum=self.maximum
        )
        self.conditional_priors = bilby.core.prior.ConditionalPriorDict(
            dict(
                var_3=self.prior_3,
                var_2=self.prior_2,
                var_0=self.prior_0,
                var_1=self.prior_1,
            )
        )
        self.conditional_priors_manually_set_items = (
            bilby.core.prior.ConditionalPriorDict()
        )
        self.test_sample = dict(var_0=0.3, var_1=0.4, var_2=0.5, var_3=0.4)
        for key, value in dict(
            var_0=self.prior_0,
            var_1=self.prior_1,
            var_2=self.prior_2,
            var_3=self.prior_3,
        ).items():
            self.conditional_priors_manually_set_items[key] = value

    def tearDown(self):
        del self.minimum
        del self.maximum
        del self.prior_0
        del self.prior_1
        del self.prior_2
        del self.prior_3
        del self.conditional_priors
        del self.conditional_priors_manually_set_items
        del self.test_sample

    def test_conditions_resolved_upon_instantiation(self):
        self.assertListEqual(
            ["var_0", "var_1", "var_2", "var_3"], self.conditional_priors.sorted_keys
        )

    def test_conditions_resolved_setting_items(self):
        self.assertListEqual(
            ["var_0", "var_1", "var_2", "var_3"],
            self.conditional_priors_manually_set_items.sorted_keys,
        )

    def test_unconditional_keys_upon_instantiation(self):
        self.assertListEqual(["var_0"], self.conditional_priors.unconditional_keys)

    def test_unconditional_keys_setting_items(self):
        self.assertListEqual(
            ["var_0"], self.conditional_priors_manually_set_items.unconditional_keys
        )

    def test_conditional_keys_upon_instantiation(self):
        self.assertListEqual(
            ["var_1", "var_2", "var_3"], self.conditional_priors.conditional_keys
        )

    def test_conditional_keys_setting_items(self):
        self.assertListEqual(
            ["var_1", "var_2", "var_3"],
            self.conditional_priors_manually_set_items.conditional_keys,
        )

    def test_prob(self):
        self.assertEqual(1, self.conditional_priors.prob(sample=self.test_sample))

    def test_prob_illegal_conditions(self):
        del self.conditional_priors["var_0"]
        with self.assertRaises(bilby.core.prior.IllegalConditionsException):
            self.conditional_priors.prob(sample=self.test_sample)

    def test_ln_prob(self):
        self.assertEqual(0, self.conditional_priors.ln_prob(sample=self.test_sample))

    def test_ln_prob_illegal_conditions(self):
        del self.conditional_priors["var_0"]
        with self.assertRaises(bilby.core.prior.IllegalConditionsException):
            self.conditional_priors.ln_prob(sample=self.test_sample)

    def test_sample_subset_all_keys(self):
        with mock.patch("numpy.random.uniform") as m:
            m.return_value = 0.5
            self.assertDictEqual(
                dict(var_0=0.5, var_1=0.5, var_2=0.5, var_3=0.5),
                self.conditional_priors.sample_subset(
                    keys=["var_0", "var_1", "var_2", "var_3"]
                ),
            )

    def test_sample_illegal_subset(self):
        with mock.patch("numpy.random.uniform") as m:
            m.return_value = 0.5
            with self.assertRaises(bilby.core.prior.IllegalConditionsException):
                self.conditional_priors.sample_subset(keys=["var_1"])

    def test_sample_multiple(self):
        def condition_func(reference_params, a):
            return dict(
                minimum=reference_params["minimum"],
                maximum=reference_params["maximum"],
                alpha=reference_params["alpha"] * a,
            )

        priors = bilby.core.prior.ConditionalPriorDict()
        priors["a"] = bilby.core.prior.Uniform(minimum=0, maximum=1)
        priors["b"] = bilby.core.prior.ConditionalPowerLaw(
            condition_func=condition_func, minimum=1, maximum=2, alpha=-2
        )
        print(priors.sample(2))

    def test_rescale(self):
        def condition_func_1_rescale(reference_parameters, var_0):
            if var_0 == 0.5:
                return dict(minimum=reference_parameters["minimum"], maximum=1)
            return reference_parameters

        def condition_func_2_rescale(reference_parameters, var_0, var_1):
            if var_0 == 0.5 and var_1 == 0.5:
                return dict(minimum=reference_parameters["minimum"], maximum=1)
            return reference_parameters

        def condition_func_3_rescale(reference_parameters, var_1, var_2):
            if var_1 == 0.5 and var_2 == 0.5:
                return dict(minimum=reference_parameters["minimum"], maximum=1)
            return reference_parameters

        self.prior_0 = bilby.core.prior.Uniform(minimum=self.minimum, maximum=1)
        self.prior_1 = bilby.core.prior.ConditionalUniform(
            condition_func=condition_func_1_rescale, minimum=self.minimum, maximum=2
        )
        self.prior_2 = bilby.core.prior.ConditionalUniform(
            condition_func=condition_func_2_rescale, minimum=self.minimum, maximum=2
        )
        self.prior_3 = bilby.core.prior.ConditionalUniform(
            condition_func=condition_func_3_rescale, minimum=self.minimum, maximum=2
        )
        self.conditional_priors = bilby.core.prior.ConditionalPriorDict(
            dict(
                var_3=self.prior_3,
                var_2=self.prior_2,
                var_0=self.prior_0,
                var_1=self.prior_1,
            )
        )
        ref_variables = [0.5, 0.5, 0.5, 0.5]
        res = self.conditional_priors.rescale(
            keys=list(self.test_sample.keys()), theta=ref_variables
        )
        self.assertListEqual(ref_variables, res)

    def test_rescale_illegal_conditions(self):
        del self.conditional_priors["var_0"]
        with self.assertRaises(bilby.core.prior.IllegalConditionsException):
            self.conditional_priors.rescale(
                keys=list(self.test_sample.keys()),
                theta=list(self.test_sample.values()),
            )

    def test_combined_conditions(self):
        def d_condition_func(reference_params, a, b, c):
            return dict(
                minimum=reference_params["minimum"], maximum=reference_params["maximum"]
            )

        def a_condition_func(reference_params, b, c):
            return dict(
                minimum=reference_params["minimum"], maximum=reference_params["maximum"]
            )

        priors = bilby.core.prior.ConditionalPriorDict()

        priors["a"] = bilby.core.prior.ConditionalUniform(
            condition_func=a_condition_func, minimum=0, maximum=1
        )

        priors["b"] = bilby.core.prior.LogUniform(minimum=1, maximum=10)

        priors["d"] = bilby.core.prior.ConditionalUniform(
            condition_func=d_condition_func, minimum=0.0, maximum=1.0
        )

        priors["c"] = bilby.core.prior.LogUniform(minimum=1, maximum=10)
        priors.sample()
        res = priors.rescale(["a", "b", "d", "c"], [0.5, 0.5, 0.5, 0.5])
        print(res)


if __name__ == "__main__":
    unittest.main()



""" A script to verify that the .AUTHOR.md file is up to date """

import re
import subprocess

special_cases = ["plasky", "thomas", "mj-will"]
AUTHORS_list = []
with open("AUTHORS.md", "r") as f:
    AUTHORS_list = " ".join([line for line in f]).lower()


lines = subprocess.check_output(["git", "shortlog", "HEAD", "-sn"]).decode("utf-8").split("\n")

if len(lines) == 0:
    raise Exception("No authors to check against")

fail_test = False
for line in lines:
    line = line.replace(".", " ")
    line = re.sub('([A-Z][a-z]+)', r' \1', re.sub('([A-Z]+)', r' \1', line))
    for element in line.split()[1:]:
        element = element.lower()
        if element not in AUTHORS_list and element not in special_cases:
            print(f"Failure: {element} not in AUTHOR.md")
            fail_test += True

if fail_test:
    raise Exception("Author check list failed.. have you added your name to the .AUTHOR file?")



import os
import shutil
import distutils.dir_util
import signal
import time
import datetime
import sys

import numpy as np
import pandas as pd

from ..utils import check_directory_exists_and_if_not_mkdir, logger
from .base_sampler import NestedSampler


class _DNest4Model(object):

    def __init__(self, log_likelihood_func, from_prior_func, widths, centers, highs, lows):
        """Initialize the DNest4 model.
        Args:
            log_likelihood_func: function
                The loglikelihood function to use during the Nested Sampling run.
            from_prior_func: function
                The function to use when randomly selecting parameter vectors from the prior space.
            widths: array_like
                The approximate widths of the prior distrbutions.
            centers: array_like
                The approximate center points of the prior distributions.
        """
        self._log_likelihood = log_likelihood_func
        self._from_prior = from_prior_func
        self._widths = widths
        self._centers = centers
        self._highs = highs
        self._lows = lows
        self._n_dim = len(widths)
        return

    def log_likelihood(self, coords):
        """The model's log_likelihood function"""
        return self._log_likelihood(coords)

    def from_prior(self):
        """The model's function to select random points from the prior space."""
        return self._from_prior()

    def perturb(self, coords):
        """The perturb function to perform Monte Carlo trial moves."""
        idx = np.random.randint(self._n_dim)

        coords[idx] += (self._widths[idx] * (np.random.uniform(size=1) - 0.5))
        cw = self._widths[idx]
        cc = self._centers[idx]

        coords[idx] = self.wrap(coords[idx], (cc - 0.5 * cw), cc + 0.5 * cw)

        return 0.0

    @staticmethod
    def wrap(x, minimum, maximum):
        if maximum <= minimum:
            raise ValueError("maximum {} <= minimum {}, when trying to wrap coordinates".format(maximum, minimum))
        return (x - minimum) % (maximum - minimum) + minimum


class DNest4(NestedSampler):

    """
    Bilby wrapper of DNest4

    Parameters
    ==========
    TBD

    Other Parameters
    ------==========
    num_particles: int
        The number of points to use in the Nested Sampling active population.
    max_num_levels: int
        The max number of diffusive likelihood levels that DNest4 should initialize
        during the Diffusive Nested Sampling run.
    backend: str
        The python DNest4 backend for storing the output.
        Options are: 'memory' and 'csv'. If 'memory' the
        DNest4 outputs are stored in memory during the run. If 'csv' the
        DNest4 outputs are written out to files with a CSV format during
        the run.
        CSV backend may not be functional right now (October 2020)
    num_steps: int
        The number of MCMC iterations to run
    new_level_interval: int
        The number of moves to run before creating a new diffusive likelihood level
    lam: float
        Set the backtracking scale length
    beta: float
        Set the strength of effect to force the histogram to equal bin counts
    seed: int
        Set the seed for the C++ random number generator
    verbose: Bool
        If True, prints information during run
    """

    default_kwargs = dict(max_num_levels=20, num_steps=500,
                          new_level_interval=10000, num_per_step=10000,
                          thread_steps=1, num_particles=1000, lam=10.0,
                          beta=100, seed=None, verbose=True, outputfiles_basename=None,
                          backend='memory')

    def __init__(self, likelihood, priors, outdir="outdir", label="label", use_ratio=False, plot=False,
                 exit_code=77, skip_import_verification=False, temporary_directory=True, **kwargs):
        super(DNest4, self).__init__(
            likelihood=likelihood, priors=priors, outdir=outdir, label=label,
            use_ratio=use_ratio, plot=plot, skip_import_verification=skip_import_verification,
            exit_code=exit_code, **kwargs)

        self.num_particles = self.kwargs["num_particles"]
        self.max_num_levels = self.kwargs["max_num_levels"]
        self._verbose = self.kwargs["verbose"]
        self._backend = self.kwargs["backend"]
        self.use_temporary_directory = temporary_directory

        self.start_time = np.nan
        self.sampler = None
        self._information = np.nan
        self._last_live_sample_info = np.nan
        self._outputfiles_basename = None
        self._temporary_outputfiles_basename = None

        signal.signal(signal.SIGTERM, self.write_current_state_and_exit)
        signal.signal(signal.SIGINT, self.write_current_state_and_exit)
        signal.signal(signal.SIGALRM, self.write_current_state_and_exit)

        # Get the estimates of the prior distributions' widths and centers.
        widths = []
        centers = []
        highs = []
        lows = []

        samples = self.priors.sample(size=10000)

        for key in self.search_parameter_keys:
            pts = samples[key]
            low = pts.min()
            high = pts.max()
            width = high - low
            center = (high + low) / 2.0
            widths.append(width)
            centers.append(center)

            highs.append(high)
            lows.append(low)

        self._widths = np.array(widths)
        self._centers = np.array(centers)
        self._highs = np.array(highs)
        self._lows = np.array(lows)

        self._dnest4_model = _DNest4Model(self.log_likelihood, self.get_random_draw_from_prior, self._widths,
                                          self._centers, self._highs, self._lows)

    def _set_backend(self):
        import dnest4
        if self._backend == 'csv':
            return dnest4.backends.CSVBackend("{}/dnest4{}/".format(self.outdir, self.label), sep=" ")
        else:
            return dnest4.backends.MemoryBackend()

    def _set_dnest4_kwargs(self):
        dnest4_keys = ["num_steps", "new_level_interval", "lam", "beta", "seed"]
        self.dnest4_kwargs = {key: self.kwargs[key] for key in dnest4_keys}

    def run_sampler(self):
        import dnest4

        self._set_dnest4_kwargs()
        backend = self._set_backend()

        self._verify_kwargs_against_default_kwargs()
        self._setup_run_directory()
        self._check_and_load_sampling_time_file()
        self.start_time = time.time()

        self.sampler = dnest4.DNest4Sampler(self._dnest4_model, backend=backend)
        out = self.sampler.sample(self.max_num_levels,
                                  num_particles=self.num_particles,
                                  **self.dnest4_kwargs)

        for i, sample in enumerate(out):
            if self._verbose and ((i + 1) % 100 == 0):
                stats = self.sampler.postprocess()
                logger.info("Iteration: {0} log(Z): {1}".format(i + 1, stats['log_Z']))

        self._calculate_and_save_sampling_time()
        self._clean_up_run_directory()

        stats = self.sampler.postprocess(resample=1)
        self.result.log_evidence = stats['log_Z']
        self._information = stats['H']
        self.result.log_evidence_err = np.sqrt(self._information / self.num_particles)

        if self._backend == 'memory':
            self._last_live_sample_info = pd.DataFrame(self.sampler.backend.sample_info[-1])
            self.result.log_likelihood_evaluations = self._last_live_sample_info['log_likelihood']
            self.result.samples = np.array(self.sampler.backend.posterior_samples)
        else:
            sample_info_path = './' + self.kwargs["outputfiles_basename"] + '/sample_info.txt'
            sample_info = np.genfromtxt(sample_info_path, comments='#', names=True)
            self.result.log_likelihood_evaluations = sample_info['log_likelihood']
            self.result.samples = np.array(self.sampler.backend.posterior_samples)

        self.result.sampler_output = out
        self.result.outputfiles_basename = self.outputfiles_basename
        self.result.sampling_time = datetime.timedelta(seconds=self.total_sampling_time)

        self.calc_likelihood_count()

        return self.result

    def _translate_kwargs(self, kwargs):
        if 'num_steps' not in kwargs:
            for equiv in self.walks_equiv_kwargs:
                if equiv in kwargs:
                    kwargs['num_steps'] = kwargs.pop(equiv)

    def _verify_kwargs_against_default_kwargs(self):
        self.outputfiles_basename = self.kwargs.pop("outputfiles_basename", None)
        super(DNest4, self)._verify_kwargs_against_default_kwargs()

    def _check_and_load_sampling_time_file(self):
        self.time_file_path = self.kwargs["outputfiles_basename"] + '/sampling_time.dat'
        if os.path.exists(self.time_file_path):
            with open(self.time_file_path, 'r') as time_file:
                self.total_sampling_time = float(time_file.readline())
        else:
            self.total_sampling_time = 0

    def _calculate_and_save_sampling_time(self):
        current_time = time.time()
        new_sampling_time = current_time - self.start_time
        self.total_sampling_time += new_sampling_time

        with open(self.time_file_path, 'w') as time_file:
            time_file.write(str(self.total_sampling_time))

        self.start_time = current_time

    def _clean_up_run_directory(self):
        if self.use_temporary_directory:
            self._move_temporary_directory_to_proper_path()
            self.kwargs["outputfiles_basename"] = self.outputfiles_basename

    @property
    def outputfiles_basename(self):
        return self._outputfiles_basename

    @outputfiles_basename.setter
    def outputfiles_basename(self, outputfiles_basename):
        if outputfiles_basename is None:
            outputfiles_basename = "{}/dnest4{}/".format(self.outdir, self.label)
        if not outputfiles_basename.endswith("/"):
            outputfiles_basename += "/"
        check_directory_exists_and_if_not_mkdir(self.outdir)
        self._outputfiles_basename = outputfiles_basename

    @property
    def temporary_outputfiles_basename(self):
        return self._temporary_outputfiles_basename

    @temporary_outputfiles_basename.setter
    def temporary_outputfiles_basename(self, temporary_outputfiles_basename):
        if not temporary_outputfiles_basename.endswith("/"):
            temporary_outputfiles_basename = "{}/".format(
                temporary_outputfiles_basename
            )
        self._temporary_outputfiles_basename = temporary_outputfiles_basename
        if os.path.exists(self.outputfiles_basename):
            shutil.copytree(
                self.outputfiles_basename, self.temporary_outputfiles_basename
            )

    def write_current_state_and_exit(self, signum=None, frame=None):
        """ Write current state and exit on exit_code """
        logger.info(
            "Run interrupted by signal {}: checkpoint and exit on {}".format(
                signum, self.exit_code
            )
        )
        self._calculate_and_save_sampling_time()
        if self.use_temporary_directory:
            self._move_temporary_directory_to_proper_path()
        sys.exit(self.exit_code)

    def _move_temporary_directory_to_proper_path(self):
        """
        Move the temporary back to the proper path

        Anything in the proper path at this point is removed including links
        """
        self._copy_temporary_directory_contents_to_proper_path()
        shutil.rmtree(self.temporary_outputfiles_basename)

    def _copy_temporary_directory_contents_to_proper_path(self):
        """
        Copy the temporary back to the proper path.
        Do not delete the temporary directory.
        """
        logger.info(
            "Overwriting {} with {}".format(
                self.outputfiles_basename, self.temporary_outputfiles_basename
            )
        )
        if self.outputfiles_basename.endswith('/'):
            outputfiles_basename_stripped = self.outputfiles_basename[:-1]
        else:
            outputfiles_basename_stripped = self.outputfiles_basename
        distutils.dir_util.copy_tree(self.temporary_outputfiles_basename, outputfiles_basename_stripped)



"""
Bilby
=====

Bilby: a user-friendly Bayesian inference library.

The aim of bilby is to provide a user-friendly interface to perform parameter
estimation. It is primarily designed and built for inference of compact
binary coalescence events in interferometric data, but it can also be used for
more general problems.

The code, and many examples are hosted at https://git.ligo.org/lscsoft/bilby.
For installation instructions see
https://lscsoft.docs.ligo.org/bilby/installation.html.

"""


import sys

from . import core, gw, hyper

from .core import utils, likelihood, prior, result, sampler
from .core.sampler import run_sampler
from .core.likelihood import Likelihood

__version__ = utils.get_version_information()


if sys.version_info < (3,):
    raise ImportError(
"""You are running bilby 0.6.4 on Python 2

Bilby 0.6.4 and above are no longer compatible with Python 2, and you still
ended up with this version installed. That's unfortunate; sorry about that.
It should not have happened. Make sure you have pip >= 9.0 to avoid this kind
of issue, as well as setuptools >= 24.2:

 $ pip install pip setuptools --upgrade

Your choices:

- Upgrade to Python 3.

- Install an older version of bilby:

 $ pip install 'bilby<0.6.4'

""")



import shutil
import unittest

from mock import MagicMock

import bilby


class TestUltranest(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.likelihood = MagicMock()
        self.priors = bilby.core.prior.PriorDict(
            dict(a=bilby.core.prior.Uniform(0, 1),
                 b=bilby.core.prior.Uniform(0, 1)))
        self.priors["a"] = bilby.core.prior.Prior(boundary="periodic")
        self.priors["b"] = bilby.core.prior.Prior(boundary="reflective")
        self.sampler = bilby.core.sampler.Ultranest(self.likelihood, self.priors,
                                                    outdir="outdir", label="label",
                                                    use_ratio=False, plot=False,
                                                    skip_import_verification=True)

    def tearDown(self):
        del self.likelihood
        del self.priors
        del self.sampler
        shutil.rmtree("outdir")

    def test_default_kwargs(self):
        expected = dict(
            resume=True,
            show_status=True,
            num_live_points=None,
            wrapped_params=None,
            derived_param_names=None,
            run_num=None,
            vectorized=False,
            num_test_samples=2,
            draw_multiple=True,
            num_bootstraps=30,
            update_interval_iter=None,
            update_interval_ncall=None,
            log_interval=None,
            dlogz=None,
            max_iters=None,
            update_interval_volume_fraction=0.2,
            viz_callback=None,
            dKL=0.5,
            frac_remain=0.01,
            Lepsilon=0.001,
            min_ess=400,
            max_ncalls=None,
            max_num_improvement_loops=-1,
            min_num_live_points=400,
            cluster_num_live_points=40,
            step_sampler=None,
        )
        self.assertListEqual([1, 0], self.sampler.kwargs["wrapped_params"])  # Check this separately
        self.sampler.kwargs["wrapped_params"] = None  # The dict comparison can't handle lists
        self.sampler.kwargs["derived_param_names"] = None
        self.sampler.kwargs["viz_callback"] = None
        self.assertDictEqual(expected, self.sampler.kwargs)

    def test_translate_kwargs(self):
        expected = dict(
            resume=True,
            show_status=True,
            num_live_points=123,
            wrapped_params=None,
            derived_param_names=None,
            run_num=None,
            vectorized=False,
            num_test_samples=2,
            draw_multiple=True,
            num_bootstraps=30,
            update_interval_iter=None,
            update_interval_ncall=None,
            log_interval=None,
            dlogz=None,
            max_iters=None,
            update_interval_volume_fraction=0.2,
            viz_callback=None,
            dKL=0.5,
            frac_remain=0.01,
            Lepsilon=0.001,
            min_ess=400,
            max_ncalls=None,
            max_num_improvement_loops=-1,
            min_num_live_points=400,
            cluster_num_live_points=40,
            step_sampler=None,
        )
        for equiv in bilby.core.sampler.base_sampler.NestedSampler.npoints_equiv_kwargs:
            new_kwargs = self.sampler.kwargs.copy()
            del new_kwargs['num_live_points']
            new_kwargs[equiv] = 123
            self.sampler.kwargs = new_kwargs
            self.sampler.kwargs["wrapped_params"] = None
            self.sampler.kwargs["derived_param_names"] = None
            self.sampler.kwargs["viz_callback"] = None
            self.assertDictEqual(expected, self.sampler.kwargs)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

import bilby
from bilby.gw import conversion


class TestBasicConversions(unittest.TestCase):
    def setUp(self):
        self.mass_1 = 1.4
        self.mass_2 = 1.3
        self.mass_ratio = 13 / 14
        self.total_mass = 2.7
        self.chirp_mass = (1.4 * 1.3) ** 0.6 / 2.7 ** 0.2
        self.symmetric_mass_ratio = (1.4 * 1.3) / 2.7 ** 2
        self.cos_angle = -1
        self.angle = np.pi
        self.lambda_1 = 300
        self.lambda_2 = 300 * (14 / 13) ** 5
        self.lambda_tilde = (
            8
            / 13
            * (
                (
                    1
                    + 7 * self.symmetric_mass_ratio
                    - 31 * self.symmetric_mass_ratio ** 2
                )
                * (self.lambda_1 + self.lambda_2)
                + (1 - 4 * self.symmetric_mass_ratio) ** 0.5
                * (
                    1
                    + 9 * self.symmetric_mass_ratio
                    - 11 * self.symmetric_mass_ratio ** 2
                )
                * (self.lambda_1 - self.lambda_2)
            )
        )
        self.delta_lambda_tilde = (
            1
            / 2
            * (
                (1 - 4 * self.symmetric_mass_ratio) ** 0.5
                * (
                    1
                    - 13272 / 1319 * self.symmetric_mass_ratio
                    + 8944 / 1319 * self.symmetric_mass_ratio ** 2
                )
                * (self.lambda_1 + self.lambda_2)
                + (
                    1
                    - 15910 / 1319 * self.symmetric_mass_ratio
                    + 32850 / 1319 * self.symmetric_mass_ratio ** 2
                    + 3380 / 1319 * self.symmetric_mass_ratio ** 3
                )
                * (self.lambda_1 - self.lambda_2)
            )
        )

    def tearDown(self):
        del self.mass_1
        del self.mass_2
        del self.mass_ratio
        del self.total_mass
        del self.chirp_mass
        del self.symmetric_mass_ratio

    def test_total_mass_and_mass_ratio_to_component_masses(self):
        mass_1, mass_2 = conversion.total_mass_and_mass_ratio_to_component_masses(
            self.mass_ratio, self.total_mass
        )
        self.assertTrue(
            all([abs(mass_1 - self.mass_1) < 1e-5, abs(mass_2 - self.mass_2) < 1e-5])
        )

    def test_chirp_mass_and_primary_mass_to_mass_ratio(self):
        mass_ratio = conversion.chirp_mass_and_primary_mass_to_mass_ratio(
            self.chirp_mass, self.mass_1
        )
        self.assertAlmostEqual(self.mass_ratio, mass_ratio)

    def test_symmetric_mass_ratio_to_mass_ratio(self):
        mass_ratio = conversion.symmetric_mass_ratio_to_mass_ratio(
            self.symmetric_mass_ratio
        )
        self.assertAlmostEqual(self.mass_ratio, mass_ratio)

    def test_chirp_mass_and_total_mass_to_symmetric_mass_ratio(self):
        symmetric_mass_ratio = conversion.chirp_mass_and_total_mass_to_symmetric_mass_ratio(
            self.chirp_mass, self.total_mass
        )
        self.assertAlmostEqual(self.symmetric_mass_ratio, symmetric_mass_ratio)

    def test_chirp_mass_and_mass_ratio_to_total_mass(self):
        total_mass = conversion.chirp_mass_and_mass_ratio_to_total_mass(
            self.chirp_mass, self.mass_ratio
        )
        self.assertAlmostEqual(self.total_mass, total_mass)

    def test_component_masses_to_chirp_mass(self):
        chirp_mass = conversion.component_masses_to_chirp_mass(self.mass_1, self.mass_2)
        self.assertAlmostEqual(self.chirp_mass, chirp_mass)

    def test_component_masses_to_total_mass(self):
        total_mass = conversion.component_masses_to_total_mass(self.mass_1, self.mass_2)
        self.assertAlmostEqual(self.total_mass, total_mass)

    def test_component_masses_to_symmetric_mass_ratio(self):
        symmetric_mass_ratio = conversion.component_masses_to_symmetric_mass_ratio(
            self.mass_1, self.mass_2
        )
        self.assertAlmostEqual(self.symmetric_mass_ratio, symmetric_mass_ratio)

    def test_component_masses_to_mass_ratio(self):
        mass_ratio = conversion.component_masses_to_mass_ratio(self.mass_1, self.mass_2)
        self.assertAlmostEqual(self.mass_ratio, mass_ratio)

    def test_mass_1_and_chirp_mass_to_mass_ratio(self):
        mass_ratio = conversion.mass_1_and_chirp_mass_to_mass_ratio(
            self.mass_1, self.chirp_mass
        )
        self.assertAlmostEqual(self.mass_ratio, mass_ratio)

    def test_lambda_tilde_to_lambda_1_lambda_2(self):
        lambda_1, lambda_2 = conversion.lambda_tilde_to_lambda_1_lambda_2(
            self.lambda_tilde, self.mass_1, self.mass_2
        )
        self.assertTrue(
            all(
                [
                    abs(self.lambda_1 - lambda_1) < 1e-5,
                    abs(self.lambda_2 - lambda_2) < 1e-5,
                ]
            )
        )

    def test_lambda_tilde_delta_lambda_tilde_to_lambda_1_lambda_2(self):
        (
            lambda_1,
            lambda_2,
        ) = conversion.lambda_tilde_delta_lambda_tilde_to_lambda_1_lambda_2(
            self.lambda_tilde, self.delta_lambda_tilde, self.mass_1, self.mass_2
        )
        self.assertTrue(
            all(
                [
                    abs(self.lambda_1 - lambda_1) < 1e-5,
                    abs(self.lambda_2 - lambda_2) < 1e-5,
                ]
            )
        )

    def test_lambda_1_lambda_2_to_lambda_tilde(self):
        lambda_tilde = conversion.lambda_1_lambda_2_to_lambda_tilde(
            self.lambda_1, self.lambda_2, self.mass_1, self.mass_2
        )
        self.assertTrue((self.lambda_tilde - lambda_tilde) < 1e-5)

    def test_lambda_1_lambda_2_to_delta_lambda_tilde(self):
        delta_lambda_tilde = conversion.lambda_1_lambda_2_to_delta_lambda_tilde(
            self.lambda_1, self.lambda_2, self.mass_1, self.mass_2
        )
        self.assertTrue((self.delta_lambda_tilde - delta_lambda_tilde) < 1e-5)


class TestConvertToLALParams(unittest.TestCase):
    def setUp(self):
        self.search_keys = []
        self.parameters = dict()
        self.component_mass_pars = dict(mass_1=1.4, mass_2=1.4)
        self.mass_parameters = self.component_mass_pars.copy()
        self.mass_parameters["mass_ratio"] = conversion.component_masses_to_mass_ratio(
            **self.component_mass_pars
        )
        self.mass_parameters[
            "symmetric_mass_ratio"
        ] = conversion.component_masses_to_symmetric_mass_ratio(
            **self.component_mass_pars
        )
        self.mass_parameters["chirp_mass"] = conversion.component_masses_to_chirp_mass(
            **self.component_mass_pars
        )
        self.mass_parameters["total_mass"] = conversion.component_masses_to_total_mass(
            **self.component_mass_pars
        )
        self.component_tidal_parameters = dict(lambda_1=300, lambda_2=300)
        self.all_component_pars = self.component_tidal_parameters.copy()
        self.all_component_pars.update(self.component_mass_pars)
        self.tidal_parameters = self.component_tidal_parameters.copy()
        self.tidal_parameters[
            "lambda_tilde"
        ] = conversion.lambda_1_lambda_2_to_lambda_tilde(**self.all_component_pars)
        self.tidal_parameters[
            "delta_lambda_tilde"
        ] = conversion.lambda_1_lambda_2_to_delta_lambda_tilde(
            **self.all_component_pars
        )

    def tearDown(self):
        del self.search_keys
        del self.parameters

    def bbh_convert(self):
        (
            self.parameters,
            self.added_keys,
        ) = conversion.convert_to_lal_binary_black_hole_parameters(self.parameters)

    def bns_convert(self):
        (
            self.parameters,
            self.added_keys,
        ) = conversion.convert_to_lal_binary_neutron_star_parameters(self.parameters)

    def test_redshift_to_luminosity_distance(self):
        self.parameters["redshift"] = 1
        dl = conversion.redshift_to_luminosity_distance(self.parameters["redshift"])
        self.bbh_convert()
        self.assertEqual(self.parameters["luminosity_distance"], dl)

    def test_comoving_to_luminosity_distance(self):
        self.parameters["comoving_distance"] = 1
        dl = conversion.comoving_distance_to_luminosity_distance(
            self.parameters["comoving_distance"]
        )
        self.bbh_convert()
        self.assertEqual(self.parameters["luminosity_distance"], dl)

    def test_source_to_lab_frame(self):
        self.parameters["test_source"] = 1
        self.parameters["redshift"] = 1
        lab = self.parameters["test_source"] * (1 + self.parameters["redshift"])
        self.bbh_convert()
        self.assertEqual(self.parameters["test"], lab)

    def _conversion_to_component_mass(self, keys):
        for key in keys:
            self.parameters[key] = self.mass_parameters[key]
        self.bbh_convert()
        self.assertAlmostEqual(
            max(
                [
                    abs(self.parameters[key] - self.component_mass_pars[key])
                    for key in ["mass_1", "mass_2"]
                ]
            ),
            0,
        )

    def test_chirp_mass_total_mass(self):
        self._conversion_to_component_mass(["chirp_mass", "total_mass"])

    def test_chirp_mass_sym_mass_ratio(self):
        self._conversion_to_component_mass(["chirp_mass", "symmetric_mass_ratio"])

    def test_chirp_mass_mass_ratio(self):
        self._conversion_to_component_mass(["chirp_mass", "mass_ratio"])

    def test_total_mass_sym_mass_ratio(self):
        self._conversion_to_component_mass(["total_mass", "symmetric_mass_ratio"])

    def test_total_mass_mass_ratio(self):
        self._conversion_to_component_mass(["total_mass", "mass_ratio"])

    def test_total_mass_mass_1(self):
        self._conversion_to_component_mass(["total_mass", "mass_1"])

    def test_total_mass_mass_2(self):
        self._conversion_to_component_mass(["total_mass", "mass_2"])

    def test_sym_mass_ratio_mass_1(self):
        self._conversion_to_component_mass(["symmetric_mass_ratio", "mass_1"])

    def test_sym_mass_ratio_mass_2(self):
        self._conversion_to_component_mass(["symmetric_mass_ratio", "mass_2"])

    def test_mass_ratio_mass_1(self):
        self._conversion_to_component_mass(["mass_ratio", "mass_1"])

    def test_mass_ratio_mass_2(self):
        self._conversion_to_component_mass(["mass_ratio", "mass_2"])

    def test_bbh_aligned_spin_to_spherical(self):
        self.parameters["chi_1"] = -0.5
        a_1 = abs(self.parameters["chi_1"])
        tilt_1 = np.arccos(np.sign(self.parameters["chi_1"]))
        phi_jl = 0.0
        phi_12 = 0.0
        self.bbh_convert()
        self.assertDictEqual(
            {
                key: self.parameters[key]
                for key in ["a_1", "tilt_1", "phi_12", "phi_jl"]
            },
            dict(a_1=a_1, tilt_1=tilt_1, phi_jl=phi_jl, phi_12=phi_12),
        )

    def test_bbh_cos_angle_to_angle_conversion(self):
        self.parameters["cos_tilt_1"] = 1
        t1 = np.arccos(self.parameters["cos_tilt_1"])
        self.bbh_convert()
        self.assertEqual(self.parameters["tilt_1"], t1)

    def _conversion_to_component_tidal(self, keys):
        for key in keys:
            self.parameters[key] = self.tidal_parameters[key]
        for key in ["mass_1", "mass_2"]:
            self.parameters[key] = self.mass_parameters[key]
        self.bns_convert()
        component_dict = {key: self.parameters[key] for key in ["lambda_1", "lambda_2"]}
        self.assertDictEqual(component_dict, self.component_tidal_parameters)

    def test_lambda_tilde_delta_lambda_tilde(self):
        self._conversion_to_component_tidal(["lambda_tilde", "delta_lambda_tilde"])

    def test_lambda_tilde(self):
        self._conversion_to_component_tidal(["lambda_tilde"])

    def test_lambda_1(self):
        self._conversion_to_component_tidal(["lambda_1"])


class TestGenerateAllParameters(unittest.TestCase):
    def setUp(self):
        self.parameters = dict(
            mass_1=36.0,
            mass_2=29.0,
            a_1=0.4,
            a_2=0.3,
            tilt_1=0.5,
            tilt_2=1.0,
            phi_12=1.7,
            phi_jl=0.3,
            luminosity_distance=2000.0,
            theta_jn=0.4,
            psi=2.659,
            phase=1.3,
            geocent_time=1126259642.413,
            ra=1.375,
            dec=-1.2108,
            lambda_tilde=1000,
            delta_lambda_tilde=0,
        )
        self.expected_bbh_keys = [
            "mass_1",
            "mass_2",
            "a_1",
            "a_2",
            "tilt_1",
            "tilt_2",
            "phi_12",
            "phi_jl",
            "luminosity_distance",
            "theta_jn",
            "psi",
            "phase",
            "geocent_time",
            "ra",
            "dec",
            "reference_frequency",
            "waveform_approximant",
            "minimum_frequency",
            "chirp_mass",
            "total_mass",
            "symmetric_mass_ratio",
            "mass_ratio",
            "iota",
            "spin_1x",
            "spin_1y",
            "spin_1z",
            "spin_2x",
            "spin_2y",
            "spin_2z",
            "phi_1",
            "phi_2",
            "chi_eff",
            "chi_1_in_plane",
            "chi_2_in_plane",
            "chi_p",
            "cos_tilt_1",
            "cos_tilt_2",
            "redshift",
            "comoving_distance",
            "mass_1_source",
            "mass_2_source",
            "chirp_mass_source",
            "total_mass_source",
        ]
        self.expected_tidal_keys = [
            "lambda_1",
            "lambda_2",
            "lambda_tilde",
            "delta_lambda_tilde",
        ]

    def test_generate_all_bbh_parameters(self):
        new_parameters = bilby.gw.conversion.generate_all_bbh_parameters(
            self.parameters
        )
        for key in self.expected_bbh_keys:
            self.assertIn(key, new_parameters)

    def test_generate_all_bns_parameters(self):
        new_parameters = bilby.gw.conversion.generate_all_bns_parameters(
            self.parameters
        )
        for key in self.expected_bbh_keys + self.expected_tidal_keys:
            self.assertIn(key, new_parameters)


class TestDistanceTransformations(unittest.TestCase):
    def setUp(self):
        self.distances = np.linspace(1, 1000, 100)

    def test_luminosity_redshift_with_cosmology(self):
        z = conversion.luminosity_distance_to_redshift(
            self.distances, cosmology="WMAP9"
        )
        dl = conversion.redshift_to_luminosity_distance(z, cosmology="WMAP9")
        self.assertAlmostEqual(max(abs(dl - self.distances)), 0, 4)

    def test_comoving_redshift_with_cosmology(self):
        z = conversion.comoving_distance_to_redshift(self.distances, cosmology="WMAP9")
        dc = conversion.redshift_to_comoving_distance(z, cosmology="WMAP9")
        self.assertAlmostEqual(max(abs(dc - self.distances)), 0, 4)

    def test_comoving_luminosity_with_cosmology(self):
        dc = conversion.comoving_distance_to_luminosity_distance(
            self.distances, cosmology="WMAP9"
        )
        dl = conversion.luminosity_distance_to_comoving_distance(dc, cosmology="WMAP9")
        self.assertAlmostEqual(max(abs(dl - self.distances)), 0, 4)


if __name__ == "__main__":
    unittest.main()



from bilby.gw import calibration
import unittest
import numpy as np


class TestBaseClass(unittest.TestCase):
    def setUp(self):
        self.model = calibration.Recalibrate()

    def tearDown(self):
        del self.model

    def test_repr(self):
        expected = "Recalibrate(prefix={})".format("'recalib_'")
        actual = repr(self.model)
        self.assertEqual(expected, actual)

    def test_calibration_factor(self):
        frequency_array = np.linspace(20, 1024, 1000)
        cal_factor = self.model.get_calibration_factor(frequency_array)
        assert np.alltrue(cal_factor.real == np.ones_like(frequency_array))


class TestCubicSpline(unittest.TestCase):
    def setUp(self):
        self.prefix = "recalib_"
        self.minimum_frequency = 20
        self.maximum_frequency = 1024
        self.n_points = 5
        self.model = calibration.CubicSpline(
            prefix=self.prefix,
            minimum_frequency=self.minimum_frequency,
            maximum_frequency=self.maximum_frequency,
            n_points=self.n_points,
        )
        self.parameters = {
            "recalib_{}_{}".format(param, ii): 0.0
            for ii in range(5)
            for param in ["amplitude", "phase"]
        }

    def tearDown(self):
        del self.prefix
        del self.minimum_frequency
        del self.maximum_frequency
        del self.n_points
        del self.model
        del self.parameters

    def test_calibration_factor(self):
        frequency_array = np.linspace(20, 1024, 1000)
        cal_factor = self.model.get_calibration_factor(
            frequency_array, **self.parameters
        )
        assert np.alltrue(cal_factor.real == np.ones_like(frequency_array))

    def test_repr(self):
        expected = "CubicSpline(prefix='{}', minimum_frequency={}, maximum_frequency={}, n_points={})".format(
            self.prefix, self.minimum_frequency, self.maximum_frequency, self.n_points
        )
        actual = repr(self.model)
        self.assertEqual(expected, actual)


class TestCubicSplineRequiresFourNodes(unittest.TestCase):
    def test_cannot_instantiate_with_too_few_nodes(self):
        for ii in range(6):
            if ii < 4:
                with self.assertRaises(ValueError):
                    calibration.CubicSpline("test", 1, 10, ii)
            else:
                calibration.CubicSpline("test", 1, 10, ii)


if __name__ == "__main__":
    unittest.main()



import unittest
import os
from shutil import rmtree

import numpy as np
import gwpy
import lal
import lalsimulation as lalsim
from scipy.stats import ks_2samp

import bilby
from bilby.gw import utils as gwutils


class TestGWUtils(unittest.TestCase):
    def setUp(self):
        self.outdir = "outdir"
        bilby.core.utils.check_directory_exists_and_if_not_mkdir(self.outdir)

    def tearDown(self):
        try:
            rmtree(self.outdir)
        except FileNotFoundError:
            pass

    def test_asd_from_freq_series(self):
        freq_data = np.array([1, 2, 3])
        df = 0.1
        asd = gwutils.asd_from_freq_series(freq_data, df)
        self.assertTrue(np.all(asd == freq_data * 2 * df ** 0.5))

    def test_psd_from_freq_series(self):
        freq_data = np.array([1, 2, 3])
        df = 0.1
        psd = gwutils.psd_from_freq_series(freq_data, df)
        self.assertTrue(np.all(psd == (freq_data * 2 * df ** 0.5) ** 2))

    def test_time_delay_from_geocenter(self):
        """
        The difference in the two detector case is due to rounding error.
        Different hardware gives different numbers in the last decimal place.
        """
        det1 = np.array([0.1, 0.2, 0.3])
        det2 = np.array([0.1, 0.2, 0.5])
        ra = 0.5
        dec = 0.2
        time = 10
        self.assertEqual(gwutils.time_delay_geocentric(det1, det1, ra, dec, time), 0)
        self.assertAlmostEqual(
            gwutils.time_delay_geocentric(det1, det2, ra, dec, time),
            1.3253791114055397e-10,
            14,
        )

    def test_get_polarization_tensor(self):
        ra = 1
        dec = 2.0
        time = 10
        psi = 0.1
        for mode in ["plus", "cross", "breathing", "longitudinal", "x", "y"]:
            p = gwutils.get_polarization_tensor(ra, dec, time, psi, mode)
            self.assertEqual(p.shape, (3, 3))
        with self.assertRaises(ValueError):
            gwutils.get_polarization_tensor(ra, dec, time, psi, "not-a-mode")

    def test_inner_product(self):
        aa = np.array([1, 2, 3])
        bb = np.array([5, 6, 7])
        frequency = np.array([0.2, 0.4, 0.6])
        PSD = bilby.gw.detector.PowerSpectralDensity.from_aligo()
        ip = gwutils.inner_product(aa, bb, frequency, PSD)
        self.assertEqual(ip, 0)

    def test_noise_weighted_inner_product(self):
        aa = np.array([1e-23, 2e-23, 3e-23])
        bb = np.array([5e-23, 6e-23, 7e-23])
        frequency = np.array([100, 101, 102])
        PSD = bilby.gw.detector.PowerSpectralDensity.from_aligo()
        psd = PSD.power_spectral_density_interpolated(frequency)
        duration = 4
        nwip = gwutils.noise_weighted_inner_product(aa, bb, psd, duration)
        self.assertEqual(nwip, 239.87768033598326)

        self.assertEqual(
            gwutils.optimal_snr_squared(aa, psd, duration),
            gwutils.noise_weighted_inner_product(aa, aa, psd, duration),
        )

    def test_matched_filter_snr(self):
        signal = np.array([1e-23, 2e-23, 3e-23])
        frequency_domain_strain = np.array([5e-23, 6e-23, 7e-23])
        frequency = np.array([100, 101, 102])
        PSD = bilby.gw.detector.PowerSpectralDensity.from_aligo()
        psd = PSD.power_spectral_density_interpolated(frequency)
        duration = 4

        mfsnr = gwutils.matched_filter_snr(
            signal, frequency_domain_strain, psd, duration
        )
        self.assertEqual(mfsnr, 25.510869054168282)

    def test_get_event_time(self):
        events = [
            "GW150914",
            "GW151012",
            "GW151226",
            "GW170104",
            "GW170608",
            "GW170729",
            "GW170809",
            "GW170814",
            "GW170817",
            "GW170818",
            "GW170823",
        ]
        for event in events:
            self.assertTrue(isinstance(gwutils.get_event_time(event), float))

        self.assertTrue(gwutils.get_event_time("GW010290") is None)

    def test_read_frame_file(self):
        start_time = 0
        end_time = 10
        channel = "H1:GDS-CALIB_STRAIN"
        N = 100
        times = np.linspace(start_time, end_time, N)
        data = np.random.normal(0, 1, N)
        ts = gwpy.timeseries.TimeSeries(data=data, times=times, t0=0)
        ts.channel = gwpy.detector.Channel(channel)
        ts.name = channel
        filename = os.path.join(self.outdir, "test.gwf")
        ts.write(filename, format="gwf")

        # Check reading without time limits
        strain = gwutils.read_frame_file(
            filename, start_time=None, end_time=None, channel=channel
        )
        self.assertEqual(strain.channel.name, channel)
        self.assertTrue(np.all(strain.value == data[:-1]))

        # Check reading with time limits
        start_cut = 2
        end_cut = 8
        strain = gwutils.read_frame_file(
            filename, start_time=start_cut, end_time=end_cut, channel=channel
        )
        idxs = (times > start_cut) & (times < end_cut)
        # Dropping the last element - for some reason gwpy drops the last element when reading in data
        self.assertTrue(np.all(strain.value == data[idxs][:-1]))

        # Check reading with unknown channels
        strain = gwutils.read_frame_file(filename, start_time=None, end_time=None)
        self.assertTrue(np.all(strain.value == data[:-1]))

        # Check reading with incorrect channel
        strain = gwutils.read_frame_file(
            filename, start_time=None, end_time=None, channel="WRONG"
        )
        self.assertTrue(np.all(strain.value == data[:-1]))

        ts = gwpy.timeseries.TimeSeries(data=data, times=times, t0=0)
        ts.name = "NOT-A-KNOWN-CHANNEL"
        ts.write(filename, format="gwf")
        strain = gwutils.read_frame_file(filename, start_time=None, end_time=None)
        self.assertEqual(strain, None)

    def test_convert_args_list_to_float(self):
        self.assertEqual(
            gwutils.convert_args_list_to_float(1, "2", 3.0), [1.0, 2.0, 3.0]
        )
        with self.assertRaises(ValueError):
            gwutils.convert_args_list_to_float(1, "2", "ten")

    def test_lalsim_SimInspiralTransformPrecessingNewInitialConditions(self):
        a = gwutils.lalsim_SimInspiralTransformPrecessingNewInitialConditions(
            0.1, 0, 0.6, 0.5, 0.6, 0.1, 0.8, 30.6, 23.2, 50, 0
        )
        self.assertTrue(len(a) == 7)

    def test_get_approximant(self):
        with self.assertRaises(ValueError):
            gwutils.lalsim_GetApproximantFromString(10)

    def test_lalsim_SimInspiralChooseFDWaveform(self):
        a = gwutils.lalsim_SimInspiralChooseFDWaveform(
            35.2,
            20.4,
            0.1,
            0.2,
            0.2,
            0.2,
            0.2,
            0.1,
            1000,
            2,
            2.3,
            45.0,
            0.1,
            10,
            0.01,
            10,
            1000,
            20,
            None,
            lalsim.IMRPhenomPv2,
        )
        self.assertEqual(len(a), 2)
        self.assertEqual(type(a[0]), lal.COMPLEX16FrequencySeries)
        self.assertEqual(type(a[1]), lal.COMPLEX16FrequencySeries)

        with self.assertRaises(RuntimeError):
            _ = gwutils.lalsim_SimInspiralChooseFDWaveform(
                35.2,
                20.4,
                0.1,
                0.2,
                0.2,
                0.2,
                0.2,
                0.1,
                1000,
                2,
                2.3,
                45.0,
                0.1,
                10,
                0.01,
                10,
                1000,
                20,
                None,
                "Fail",
            )

        with self.assertRaises(ValueError):
            _ = gwutils.lalsim_SimInspiralChooseFDWaveform(
                35.2,
                20.4,
                0.1,
                0.2,
                0.2,
                0.2,
                0.2,
                0.1,
                1000,
                2,
                2.3,
                45.0,
                0.1,
                10,
                0.01,
                10,
                1000,
                20,
                None,
                1.5,
            )


class TestSkyFrameConversion(unittest.TestCase):

    def setUp(self) -> None:
        self.priors = bilby.core.prior.PriorDict()
        self.priors["ra"] = bilby.core.prior.Uniform(0, 2 * np.pi)
        self.priors["dec"] = bilby.core.prior.Cosine()
        self.priors["azimuth"] = bilby.core.prior.Uniform(0, 2 * np.pi)
        self.priors["zenith"] = bilby.core.prior.Sine()
        self.priors["time"] = bilby.core.prior.Uniform(-0.1, 0.1)
        self.ifos = bilby.gw.detector.InterferometerList(["H1", "L1"])
        self.samples = self.priors.sample(10000)

    def tearDown(self) -> None:
        del self.priors
        del self.ifos
        del self.samples

    def test_conversion_gives_correct_prior(self) -> None:
        zeniths = self.samples["zenith"]
        azimuths = self.samples["azimuth"]
        times = self.samples["time"]
        args = zip(*[
            (zenith, azimuth, time, self.ifos)
            for zenith, azimuth, time in zip(zeniths, azimuths, times)
        ])
        ras, decs = zip(*map(bilby.gw.utils.zenith_azimuth_to_ra_dec, *args))
        self.assertGreaterEqual(ks_2samp(self.samples["ra"], ras).pvalue, 0.01)
        self.assertGreaterEqual(ks_2samp(self.samples["dec"], decs).pvalue, 0.01)


if __name__ == "__main__":
    unittest.main()




import numpy as np
from pandas import DataFrame

from .base_sampler import NestedSampler


class Nestle(NestedSampler):
    """bilby wrapper `nestle.Sampler` (http://kylebarbary.com/nestle/)

    All positional and keyword arguments (i.e., the args and kwargs) passed to
    `run_sampler` will be propagated to `nestle.sample`, see documentation for
    that function for further help. Under Other Parameters, we list commonly
    used kwargs and the bilby defaults

    Parameters
    ==========
    npoints: int
        The number of live points, note this can also equivalently be given as
        one of [nlive, nlives, n_live_points]
    method: {'classic', 'single', 'multi'} ('multi')
        Method used to select new points
    verbose: Bool
        If true, print information information about the convergence during
        sampling

    """
    default_kwargs = dict(verbose=True, method='multi', npoints=500,
                          update_interval=None, npdim=None, maxiter=None,
                          maxcall=None, dlogz=None, decline_factor=None,
                          rstate=None, callback=None, steps=20, enlarge=1.2)

    def _translate_kwargs(self, kwargs):
        if 'npoints' not in kwargs:
            for equiv in self.npoints_equiv_kwargs:
                if equiv in kwargs:
                    kwargs['npoints'] = kwargs.pop(equiv)
        if 'steps' not in kwargs:
            for equiv in self.walks_equiv_kwargs:
                if equiv in kwargs:
                    kwargs['steps'] = kwargs.pop(equiv)

    def _verify_kwargs_against_default_kwargs(self):
        if self.kwargs['verbose']:
            import nestle
            self.kwargs['callback'] = nestle.print_progress
            self.kwargs.pop('verbose')
        NestedSampler._verify_kwargs_against_default_kwargs(self)

    def run_sampler(self):
        """ Runs Nestle sampler with given kwargs and returns the result

        Returns
        =======
        bilby.core.result.Result: Packaged information about the result

        """
        import nestle
        out = nestle.sample(
            loglikelihood=self.log_likelihood,
            prior_transform=self.prior_transform,
            ndim=self.ndim, **self.kwargs)
        print("")

        self.result.sampler_output = out
        self.result.samples = nestle.resample_equal(out.samples, out.weights)
        self.result.nested_samples = DataFrame(
            out.samples, columns=self.search_parameter_keys)
        self.result.nested_samples['weights'] = out.weights
        self.result.nested_samples['log_likelihood'] = out.logl
        self.result.log_likelihood_evaluations = self.reorder_loglikelihoods(
            unsorted_loglikelihoods=out.logl, unsorted_samples=out.samples,
            sorted_samples=self.result.samples)
        self.result.log_evidence = out.logz
        self.result.log_evidence_err = out.logzerr
        self.result.information_gain = out.h
        self.calc_likelihood_count()
        return self.result

    def _run_test(self):
        """
        Runs to test whether the sampler is properly running with the given
        kwargs without actually running to the end

        Returns
        =======
        bilby.core.result.Result: Dummy container for sampling results.

        """
        import nestle
        kwargs = self.kwargs.copy()
        kwargs['maxiter'] = 2
        nestle.sample(
            loglikelihood=self.log_likelihood,
            prior_transform=self.prior_transform,
            ndim=self.ndim, **kwargs)
        self.result.samples = np.random.uniform(0, 1, (100, self.ndim))
        self.result.log_evidence = np.nan
        self.result.log_evidence_err = np.nan
        self.calc_likelihood_count()
        return self.result



import matplotlib

matplotlib.use("Agg")

import unittest
import os
import shutil
import logging

# Required to run the tests
from past.builtins import execfile
import bilby.core.utils

# Imported to ensure the examples run
import numpy as np  # noqa: F401
import inspect  # noqa: F401

bilby.core.utils.command_line_args.bilby_test_mode = True


class ExampleTest(unittest.TestCase):
    outdir = "outdir"
    dir_path = os.path.dirname(os.path.realpath(__file__))
    dir_path = os.path.abspath(os.path.join(dir_path, os.path.pardir))

    @classmethod
    def setUpClass(cls):
        if os.path.isdir(cls.outdir):
            try:
                shutil.rmtree(cls.outdir)
            except OSError:
                logging.warning("{} not removed prior to tests".format(cls.outdir))

    @classmethod
    def tearDownClass(cls):
        if os.path.isdir(cls.outdir):
            try:
                shutil.rmtree(cls.outdir)
            except OSError:
                logging.warning("{} not removed prior to tests".format(cls.outdir))

    def test_examples(self):
        """ Loop over examples to check they run """
        examples = [
            "examples/core_examples/linear_regression.py",
            "examples/core_examples/linear_regression_unknown_noise.py",
        ]
        for filename in examples:
            print("Testing {}".format(filename))
            execfile(filename)

    def test_gw_examples(self):
        """ Loop over examples to check they run """
        examples = [
            "examples/gw_examples/injection_examples/fast_tutorial.py",
            "examples/gw_examples/data_examples/GW150914.py",
        ]
        for filename in examples:
            print("Testing {}".format(filename))
            execfile(filename)


if __name__ == "__main__":
    unittest.main()



import importlib
import os
import shutil
import distutils.dir_util
import signal
import time
import datetime
import sys

import numpy as np

from ..utils import check_directory_exists_and_if_not_mkdir
from ..utils import logger
from .base_sampler import NestedSampler


class Pymultinest(NestedSampler):
    """
    bilby wrapper of pymultinest
    (https://github.com/JohannesBuchner/PyMultiNest)

    All positional and keyword arguments (i.e., the args and kwargs) passed to
    `run_sampler` will be propagated to `pymultinest.run`, see documentation
    for that class for further help. Under Other Parameters, we list commonly
    used kwargs and the bilby defaults.

    Parameters
    ==========
    npoints: int
        The number of live points, note this can also equivalently be given as
        one of [nlive, nlives, n_live_points]
    importance_nested_sampling: bool, (False)
        If true, use importance nested sampling
    sampling_efficiency: float or {'parameter', 'model'}, ('parameter')
        Defines the sampling efficiency
    verbose: Bool
        If true, print information information about the convergence during
    resume: bool
        If true, resume run from checkpoint (if available)

    """

    default_kwargs = dict(
        importance_nested_sampling=False,
        resume=True,
        verbose=True,
        sampling_efficiency="parameter",
        n_live_points=500,
        n_params=None,
        n_clustering_params=None,
        wrapped_params=None,
        multimodal=True,
        const_efficiency_mode=False,
        evidence_tolerance=0.5,
        n_iter_before_update=100,
        null_log_evidence=-1e90,
        max_modes=100,
        mode_tolerance=-1e90,
        outputfiles_basename=None,
        seed=-1,
        context=0,
        write_output=True,
        log_zero=-1e100,
        max_iter=0,
        init_MPI=False,
        dump_callback=None,
    )

    def __init__(
        self,
        likelihood,
        priors,
        outdir="outdir",
        label="label",
        use_ratio=False,
        plot=False,
        exit_code=77,
        skip_import_verification=False,
        temporary_directory=True,
        **kwargs
    ):
        super(Pymultinest, self).__init__(
            likelihood=likelihood,
            priors=priors,
            outdir=outdir,
            label=label,
            use_ratio=use_ratio,
            plot=plot,
            skip_import_verification=skip_import_verification,
            exit_code=exit_code,
            **kwargs
        )
        self._apply_multinest_boundaries()
        self.exit_code = exit_code
        using_mpi = len([key for key in os.environ if "MPI" in key])
        if using_mpi and temporary_directory:
            logger.info(
                "Temporary directory incompatible with MPI, "
                "will run in original directory"
            )
        self.use_temporary_directory = temporary_directory and not using_mpi

        signal.signal(signal.SIGTERM, self.write_current_state_and_exit)
        signal.signal(signal.SIGINT, self.write_current_state_and_exit)
        signal.signal(signal.SIGALRM, self.write_current_state_and_exit)

    def _translate_kwargs(self, kwargs):
        if "n_live_points" not in kwargs:
            for equiv in self.npoints_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["n_live_points"] = kwargs.pop(equiv)

    def _verify_kwargs_against_default_kwargs(self):
        """ Check the kwargs """

        self.outputfiles_basename = self.kwargs.pop("outputfiles_basename", None)

        # for PyMultiNest >=2.9 the n_params kwarg cannot be None
        if self.kwargs["n_params"] is None:
            self.kwargs["n_params"] = self.ndim
        if self.kwargs['dump_callback'] is None:
            self.kwargs['dump_callback'] = self._dump_callback
        NestedSampler._verify_kwargs_against_default_kwargs(self)

    def _dump_callback(self, *args, **kwargs):
        if self.use_temporary_directory:
            self._copy_temporary_directory_contents_to_proper_path()
        self._calculate_and_save_sampling_time()

    def _apply_multinest_boundaries(self):
        if self.kwargs["wrapped_params"] is None:
            self.kwargs["wrapped_params"] = list()
            for param in self.search_parameter_keys:
                if self.priors[param].boundary == "periodic":
                    self.kwargs["wrapped_params"].append(1)
                else:
                    self.kwargs["wrapped_params"].append(0)

    @property
    def outputfiles_basename(self):
        return self._outputfiles_basename

    @outputfiles_basename.setter
    def outputfiles_basename(self, outputfiles_basename):
        if outputfiles_basename is None:
            outputfiles_basename = "{}/pm_{}/".format(self.outdir, self.label)
        if not outputfiles_basename.endswith("/"):
            outputfiles_basename += "/"
        check_directory_exists_and_if_not_mkdir(self.outdir)
        self._outputfiles_basename = outputfiles_basename

    @property
    def temporary_outputfiles_basename(self):
        return self._temporary_outputfiles_basename

    @temporary_outputfiles_basename.setter
    def temporary_outputfiles_basename(self, temporary_outputfiles_basename):
        if not temporary_outputfiles_basename.endswith("/"):
            temporary_outputfiles_basename = "{}/".format(
                temporary_outputfiles_basename
            )
        self._temporary_outputfiles_basename = temporary_outputfiles_basename
        if os.path.exists(self.outputfiles_basename):
            shutil.copytree(
                self.outputfiles_basename, self.temporary_outputfiles_basename
            )

    def write_current_state_and_exit(self, signum=None, frame=None):
        """ Write current state and exit on exit_code """
        logger.info(
            "Run interrupted by signal {}: checkpoint and exit on {}".format(
                signum, self.exit_code
            )
        )
        self._calculate_and_save_sampling_time()
        if self.use_temporary_directory:
            self._move_temporary_directory_to_proper_path()
        sys.exit(self.exit_code)

    def _copy_temporary_directory_contents_to_proper_path(self):
        """
        Copy the temporary back to the proper path.
        Do not delete the temporary directory.
        """
        logger.info(
            "Overwriting {} with {}".format(
                self.outputfiles_basename, self.temporary_outputfiles_basename
            )
        )
        if self.outputfiles_basename.endswith('/'):
            outputfiles_basename_stripped = self.outputfiles_basename[:-1]
        else:
            outputfiles_basename_stripped = self.outputfiles_basename
        distutils.dir_util.copy_tree(self.temporary_outputfiles_basename, outputfiles_basename_stripped)

    def _move_temporary_directory_to_proper_path(self):
        """
        Copy the temporary back to the proper path

        Anything in the temporary directory at this point is removed
        """
        self._copy_temporary_directory_contents_to_proper_path()
        shutil.rmtree(self.temporary_outputfiles_basename)

    def run_sampler(self):
        import pymultinest

        self._verify_kwargs_against_default_kwargs()

        self._setup_run_directory()
        self._check_and_load_sampling_time_file()

        # Overwrite pymultinest's signal handling function
        pm_run = importlib.import_module("pymultinest.run")
        pm_run.interrupt_handler = self.write_current_state_and_exit

        self.start_time = time.time()
        out = pymultinest.solve(
            LogLikelihood=self.log_likelihood,
            Prior=self.prior_transform,
            n_dims=self.ndim,
            **self.kwargs
        )
        self._calculate_and_save_sampling_time()

        self._clean_up_run_directory()

        post_equal_weights = os.path.join(
            self.outputfiles_basename, "post_equal_weights.dat"
        )
        post_equal_weights_data = np.loadtxt(post_equal_weights)
        self.result.log_likelihood_evaluations = post_equal_weights_data[:, -1]
        self.result.sampler_output = out
        self.result.samples = post_equal_weights_data[:, :-1]
        self.result.log_evidence = out["logZ"]
        self.result.log_evidence_err = out["logZerr"]
        self.calc_likelihood_count()
        self.result.outputfiles_basename = self.outputfiles_basename
        self.result.sampling_time = datetime.timedelta(seconds=self.total_sampling_time)
        return self.result

    def _check_and_load_sampling_time_file(self):
        self.time_file_path = self.kwargs["outputfiles_basename"] + '/sampling_time.dat'
        if os.path.exists(self.time_file_path):
            with open(self.time_file_path, 'r') as time_file:
                self.total_sampling_time = float(time_file.readline())
        else:
            self.total_sampling_time = 0

    def _calculate_and_save_sampling_time(self):
        current_time = time.time()
        new_sampling_time = current_time - self.start_time
        self.total_sampling_time += new_sampling_time
        self.start_time = current_time
        with open(self.time_file_path, 'w') as time_file:
            time_file.write(str(self.total_sampling_time))

    def _clean_up_run_directory(self):
        if self.use_temporary_directory:
            self._move_temporary_directory_to_proper_path()
            self.kwargs["outputfiles_basename"] = self.outputfiles_basename



import unittest

import numpy as np
from mock import MagicMock

import bilby


class TestPolyChord(unittest.TestCase):
    def setUp(self):
        self.likelihood = MagicMock()
        self.priors = bilby.core.prior.PriorDict(
            dict(a=bilby.core.prior.Uniform(0, 1), b=bilby.core.prior.Uniform(0, 1))
        )
        self.sampler = bilby.core.sampler.PyPolyChord(
            self.likelihood,
            self.priors,
            outdir="outdir",
            label="polychord",
            use_ratio=False,
            plot=False,
            skip_import_verification=True,
        )

    def tearDown(self):
        del self.likelihood
        del self.priors
        del self.sampler

    def test_default_kwargs(self):
        expected = dict(
            use_polychord_defaults=False,
            nlive=self.sampler.ndim * 25,
            num_repeats=self.sampler.ndim * 5,
            nprior=-1,
            do_clustering=True,
            feedback=1,
            precision_criterion=0.001,
            logzero=-1e30,
            max_ndead=-1,
            boost_posterior=0.0,
            posteriors=True,
            equals=True,
            cluster_posteriors=True,
            write_resume=True,
            write_paramnames=False,
            read_resume=True,
            write_stats=True,
            write_live=True,
            write_dead=True,
            write_prior=True,
            compression_factor=np.exp(-1),
            base_dir="outdir",
            file_root="polychord",
            seed=-1,
            grade_dims=list([self.sampler.ndim]),
            grade_frac=list([1.0] * len([self.sampler.ndim])),
            nlives={},
        )
        self.sampler._setup_dynamic_defaults()
        self.assertDictEqual(expected, self.sampler.kwargs)

    def test_translate_kwargs(self):
        expected = dict(
            use_polychord_defaults=False,
            nlive=123,
            num_repeats=self.sampler.ndim * 5,
            nprior=-1,
            do_clustering=True,
            feedback=1,
            precision_criterion=0.001,
            logzero=-1e30,
            max_ndead=-1,
            boost_posterior=0.0,
            posteriors=True,
            equals=True,
            cluster_posteriors=True,
            write_resume=True,
            write_paramnames=False,
            read_resume=True,
            write_stats=True,
            write_live=True,
            write_dead=True,
            write_prior=True,
            compression_factor=np.exp(-1),
            base_dir="outdir",
            file_root="polychord",
            seed=-1,
            grade_dims=list([self.sampler.ndim]),
            grade_frac=list([1.0] * len([self.sampler.ndim])),
            nlives={},
        )
        self.sampler._setup_dynamic_defaults()
        for equiv in bilby.core.sampler.base_sampler.NestedSampler.npoints_equiv_kwargs:
            new_kwargs = self.sampler.kwargs.copy()
            del new_kwargs["nlive"]
            new_kwargs[equiv] = 123
            self.sampler.kwargs = new_kwargs
            self.assertDictEqual(expected, self.sampler.kwargs)


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import pandas as pd
import shutil
import os
import json

import bilby


class TestJson(unittest.TestCase):
    def setUp(self):
        self.encoder = bilby.core.utils.BilbyJsonEncoder
        self.decoder = bilby.core.utils.decode_bilby_json

    def test_list_encoding(self):
        data = dict(x=[1, 2, 3.4])
        encoded = json.dumps(data, cls=self.encoder)
        decoded = json.loads(encoded, object_hook=self.decoder)
        self.assertEqual(data.keys(), decoded.keys())
        self.assertEqual(type(data["x"]), type(decoded["x"]))
        self.assertTrue(np.all(data["x"] == decoded["x"]))

    def test_array_encoding(self):
        data = dict(x=np.array([1, 2, 3.4]))
        encoded = json.dumps(data, cls=self.encoder)
        decoded = json.loads(encoded, object_hook=self.decoder)
        self.assertEqual(data.keys(), decoded.keys())
        self.assertEqual(type(data["x"]), type(decoded["x"]))
        self.assertTrue(np.all(data["x"] == decoded["x"]))

    def test_complex_encoding(self):
        data = dict(x=1 + 3j)
        encoded = json.dumps(data, cls=self.encoder)
        decoded = json.loads(encoded, object_hook=self.decoder)
        self.assertEqual(data.keys(), decoded.keys())
        self.assertEqual(type(data["x"]), type(decoded["x"]))
        self.assertTrue(np.all(data["x"] == decoded["x"]))

    def test_dataframe_encoding(self):
        data = dict(data=pd.DataFrame(dict(x=[3, 4, 5], y=[5, 6, 7])))
        encoded = json.dumps(data, cls=self.encoder)
        decoded = json.loads(encoded, object_hook=self.decoder)
        self.assertEqual(data.keys(), decoded.keys())
        self.assertEqual(type(data["data"]), type(decoded["data"]))
        self.assertTrue(np.all(data["data"]["x"] == decoded["data"]["x"]))
        self.assertTrue(np.all(data["data"]["y"] == decoded["data"]["y"]))


class TestResult(unittest.TestCase):
    def setUp(self):
        np.random.seed(7)
        bilby.utils.command_line_args.bilby_test_mode = False
        priors = bilby.prior.PriorDict(
            dict(
                x=bilby.prior.Uniform(0, 1, "x", latex_label="$x$", unit="s"),
                y=bilby.prior.Uniform(0, 1, "y", latex_label="$y$", unit="m"),
                c=1,
                d=2,
            )
        )
        result = bilby.core.result.Result(
            label="label",
            outdir="outdir",
            sampler="nestle",
            search_parameter_keys=["x", "y"],
            fixed_parameter_keys=["c", "d"],
            priors=priors,
            sampler_kwargs=dict(test="test", func=lambda x: x),
            injection_parameters=dict(x=0.5, y=0.5),
            meta_data=dict(test="test"),
        )

        n = 100
        posterior = pd.DataFrame(
            dict(x=np.random.normal(0, 1, n), y=np.random.normal(0, 1, n))
        )
        result.posterior = posterior
        result.log_evidence = 10
        result.log_evidence_err = 11
        result.log_bayes_factor = 12
        result.log_noise_evidence = 13
        self.result = result
        pass

    def tearDown(self):
        bilby.utils.command_line_args.bilby_test_mode = True
        try:
            shutil.rmtree(self.result.outdir)
        except OSError:
            pass
        del self.result
        pass

    def test_result_file_name_default(self):
        outdir = "outdir"
        label = "label"
        self.assertEqual(
            bilby.core.result.result_file_name(outdir, label),
            "{}/{}_result.json".format(outdir, label),
        )

    def test_result_file_name_hdf5(self):
        outdir = "outdir"
        label = "label"
        self.assertEqual(
            bilby.core.result.result_file_name(outdir, label, extension="hdf5"),
            "{}/{}_result.hdf5".format(outdir, label),
        )

    def test_fail_save_and_load(self):
        with self.assertRaises(ValueError):
            bilby.core.result.read_in_result()

        with self.assertRaises(ValueError):
            bilby.core.result.read_in_result(filename="no_file_extension")

        with self.assertRaises(IOError):
            bilby.core.result.read_in_result(filename="not/a/file.json")

    def test_unset_priors(self):
        result = bilby.core.result.Result(
            label="label",
            outdir="outdir",
            sampler="nestle",
            search_parameter_keys=["x", "y"],
            fixed_parameter_keys=["c", "d"],
            priors=None,
            sampler_kwargs=dict(test="test"),
            injection_parameters=dict(x=0.5, y=0.5),
            meta_data=dict(test="test"),
        )
        with self.assertRaises(ValueError):
            _ = result.priors
        self.assertEqual(result.parameter_labels, result.search_parameter_keys)
        self.assertEqual(
            result.parameter_labels_with_unit, result.search_parameter_keys
        )

    def test_unknown_priors_fail(self):
        with self.assertRaises(ValueError):
            bilby.core.result.Result(
                label="label",
                outdir="outdir",
                sampler="nestle",
                search_parameter_keys=["x", "y"],
                fixed_parameter_keys=["c", "d"],
                priors=["a", "b"],
                sampler_kwargs=dict(test="test"),
                injection_parameters=dict(x=0.5, y=0.5),
                meta_data=dict(test="test"),
            )

    def test_set_samples(self):
        samples = [1, 2, 3]
        self.result.samples = samples
        self.assertEqual(samples, self.result.samples)

    def test_set_nested_samples(self):
        nested_samples = [1, 2, 3]
        self.result.nested_samples = nested_samples
        self.assertEqual(nested_samples, self.result.nested_samples)

    def test_set_walkers(self):
        walkers = [1, 2, 3]
        self.result.walkers = walkers
        self.assertEqual(walkers, self.result.walkers)

    def test_set_nburn(self):
        nburn = 1
        self.result.nburn = nburn
        self.assertEqual(nburn, self.result.nburn)

    def test_unset_posterior(self):
        self.result.posterior = None
        with self.assertRaises(ValueError):
            _ = self.result.posterior

    def test_save_and_load_hdf5(self):
        self.result.save_to_file(extension="hdf5")
        loaded_result = bilby.core.result.read_in_result(
            outdir=self.result.outdir, label=self.result.label, extension="hdf5"
        )
        self.assertTrue(
            pd.DataFrame.equals(self.result.posterior, loaded_result.posterior)
        )
        self.assertTrue(
            self.result.fixed_parameter_keys == loaded_result.fixed_parameter_keys
        )
        self.assertTrue(
            self.result.search_parameter_keys == loaded_result.search_parameter_keys
        )
        self.assertEqual(self.result.meta_data, loaded_result.meta_data)
        self.assertEqual(
            self.result.injection_parameters, loaded_result.injection_parameters
        )
        self.assertEqual(self.result.log_evidence, loaded_result.log_evidence)
        self.assertEqual(
            self.result.log_noise_evidence, loaded_result.log_noise_evidence
        )
        self.assertEqual(self.result.log_evidence_err, loaded_result.log_evidence_err)
        self.assertEqual(self.result.log_bayes_factor, loaded_result.log_bayes_factor)
        self.assertEqual(self.result.priors["x"], loaded_result.priors["x"])
        self.assertEqual(self.result.priors["y"], loaded_result.priors["y"])
        self.assertEqual(self.result.priors["c"], loaded_result.priors["c"])
        self.assertEqual(self.result.priors["d"], loaded_result.priors["d"])

    def test_save_and_load_default(self):
        self.result.save_to_file()
        loaded_result = bilby.core.result.read_in_result(
            outdir=self.result.outdir, label=self.result.label
        )
        self.assertTrue(
            np.array_equal(
                self.result.posterior.sort_values(by=["x"]),
                loaded_result.posterior.sort_values(by=["x"]),
            )
        )
        self.assertTrue(
            self.result.fixed_parameter_keys == loaded_result.fixed_parameter_keys
        )
        self.assertTrue(
            self.result.search_parameter_keys == loaded_result.search_parameter_keys
        )
        self.assertEqual(self.result.meta_data, loaded_result.meta_data)
        self.assertEqual(
            self.result.injection_parameters, loaded_result.injection_parameters
        )
        self.assertEqual(self.result.log_evidence, loaded_result.log_evidence)
        self.assertEqual(
            self.result.log_noise_evidence, loaded_result.log_noise_evidence
        )
        self.assertEqual(self.result.log_evidence_err, loaded_result.log_evidence_err)
        self.assertEqual(self.result.log_bayes_factor, loaded_result.log_bayes_factor)
        self.assertEqual(self.result.priors["x"], loaded_result.priors["x"])
        self.assertEqual(self.result.priors["y"], loaded_result.priors["y"])
        self.assertEqual(self.result.priors["c"], loaded_result.priors["c"])
        self.assertEqual(self.result.priors["d"], loaded_result.priors["d"])

    def test_save_and_load_gzip(self):
        self.result.save_to_file(gzip=True)
        loaded_result = bilby.core.result.read_in_result(
            outdir=self.result.outdir, label=self.result.label, gzip=True
        )
        self.assertTrue(
            np.array_equal(
                self.result.posterior.sort_values(by=["x"]),
                loaded_result.posterior.sort_values(by=["x"]),
            )
        )
        self.assertTrue(
            self.result.fixed_parameter_keys == loaded_result.fixed_parameter_keys
        )
        self.assertTrue(
            self.result.search_parameter_keys == loaded_result.search_parameter_keys
        )
        self.assertEqual(self.result.meta_data, loaded_result.meta_data)
        self.assertEqual(
            self.result.injection_parameters, loaded_result.injection_parameters
        )
        self.assertEqual(self.result.log_evidence, loaded_result.log_evidence)
        self.assertEqual(
            self.result.log_noise_evidence, loaded_result.log_noise_evidence
        )
        self.assertEqual(self.result.log_evidence_err, loaded_result.log_evidence_err)
        self.assertEqual(self.result.log_bayes_factor, loaded_result.log_bayes_factor)
        self.assertEqual(self.result.priors["x"], loaded_result.priors["x"])
        self.assertEqual(self.result.priors["y"], loaded_result.priors["y"])
        self.assertEqual(self.result.priors["c"], loaded_result.priors["c"])
        self.assertEqual(self.result.priors["d"], loaded_result.priors["d"])

    def test_save_and_dont_overwrite_default(self):
        shutil.rmtree(
            "{}/{}_result.json.old".format(self.result.outdir, self.result.label),
            ignore_errors=True,
        )
        self.result.save_to_file(overwrite=False)
        self.result.save_to_file(overwrite=False)
        self.assertTrue(
            os.path.isfile(
                "{}/{}_result.json.old".format(self.result.outdir, self.result.label)
            )
        )

    def test_save_and_dont_overwrite_hdf5(self):
        shutil.rmtree(
            "{}/{}_result.hdf5.old".format(self.result.outdir, self.result.label),
            ignore_errors=True,
        )
        self.result.save_to_file(overwrite=False, extension="hdf5")
        self.result.save_to_file(overwrite=False, extension="hdf5")
        self.assertTrue(
            os.path.isfile(
                "{}/{}_result.hdf5.old".format(self.result.outdir, self.result.label)
            )
        )

    def test_save_and_overwrite_hdf5(self):
        shutil.rmtree(
            "{}/{}_result.hdf5.old".format(self.result.outdir, self.result.label),
            ignore_errors=True,
        )
        self.result.save_to_file(overwrite=True, extension="hdf5")
        self.result.save_to_file(overwrite=True, extension="hdf5")
        self.assertFalse(
            os.path.isfile(
                "{}/{}_result.hdf5.old".format(self.result.outdir, self.result.label)
            )
        )

    def test_save_and_overwrite_default(self):
        shutil.rmtree(
            "{}/{}_result.json.old".format(self.result.outdir, self.result.label),
            ignore_errors=True,
        )
        self.result.save_to_file(overwrite=True, extension="hdf5")
        self.result.save_to_file(overwrite=True, extension="hdf5")
        self.assertFalse(
            os.path.isfile(
                "{}/{}_result.h5.old".format(self.result.outdir, self.result.label)
            )
        )

    def test_save_and_overwrite_default_2(self):
        shutil.rmtree(
            "{}/{}_result.json.old".format(self.result.outdir, self.result.label),
            ignore_errors=True,
        )
        self.result.save_to_file(overwrite=True)
        self.result.save_to_file(overwrite=True)
        self.assertFalse(
            os.path.isfile(
                "{}/{}_result.json.old".format(self.result.outdir, self.result.label)
            )
        )

    def test_save_samples(self):
        self.result.save_posterior_samples()
        filename = "{}/{}_posterior_samples.dat".format(
            self.result.outdir, self.result.label
        )
        self.assertTrue(os.path.isfile(filename))
        df = pd.read_csv(filename, sep=" ")
        self.assertTrue(np.allclose(self.result.posterior.values, df.values))

    def test_save_samples_from_filename(self):
        filename = "{}/{}_posterior_samples_OTHER.dat".format(
            self.result.outdir, self.result.label
        )
        self.result.save_posterior_samples(filename=filename)
        self.assertTrue(os.path.isfile(filename))
        df = pd.read_csv(filename, sep=" ")
        self.assertTrue(np.allclose(self.result.posterior.values, df.values))

    def test_save_samples_numpy_load(self):
        self.result.save_posterior_samples()
        filename = "{}/{}_posterior_samples.dat".format(
            self.result.outdir, self.result.label
        )
        self.assertTrue(os.path.isfile(filename))
        data = np.genfromtxt(filename, names=True)
        df = pd.read_csv(filename, sep=" ")
        self.assertTrue(len(data.dtype) == len(df.keys()))

    def test_samples_to_posterior_simple(self):
        self.result.posterior = None
        x = [1, 2, 3]
        y = [4, 6, 8]
        self.result.samples = np.array([x, y]).T
        self.result.samples_to_posterior()
        self.assertTrue(all(self.result.posterior["x"] == x))
        self.assertTrue(all(self.result.posterior["y"] == y))
        self.assertTrue(np.all(None == self.result.posterior.log_likelihood.values))  # noqa

    def test_samples_to_posterior(self):
        self.result.posterior = None
        x = [1, 2, 3]
        y = [4, 6, 8]
        log_likelihood = np.array([6, 7, 8])
        self.result.samples = np.array([x, y]).T
        self.result.log_likelihood_evaluations = log_likelihood
        self.result.samples_to_posterior(priors=self.result.priors)
        self.assertTrue(all(self.result.posterior["x"] == x))
        self.assertTrue(all(self.result.posterior["y"] == y))
        self.assertTrue(
            np.array_equal(self.result.posterior.log_likelihood.values, log_likelihood)
        )
        self.assertTrue(
            all(self.result.posterior.c.values == self.result.priors["c"].peak)
        )
        self.assertTrue(
            all(self.result.posterior.d.values == self.result.priors["d"].peak)
        )

    def test_calculate_prior_values(self):
        self.result.calculate_prior_values(priors=self.result.priors)
        self.assertEqual(len(self.result.posterior), len(self.result.prior_values))

    def test_plot_multiple(self):
        filename = "{}/multiple.png".format(self.result.outdir)
        bilby.core.result.plot_multiple([self.result, self.result], filename=filename)
        self.assertTrue(os.path.isfile(filename))
        os.remove(filename)

    def test_plot_walkers(self):
        self.result.walkers = np.random.uniform(0, 1, (10, 11, 2))
        self.result.nburn = 5
        self.result.plot_walkers()
        self.assertTrue(
            os.path.isfile(
                "{}/{}_walkers.png".format(self.result.outdir, self.result.label)
            )
        )

    def test_plot_with_data(self):
        x = np.linspace(0, 1, 10)
        y = np.linspace(0, 1, 10)

        def model(xx, theta):
            return xx

        self.result.posterior = pd.DataFrame(dict(theta=[1, 2, 3]))
        self.result.plot_with_data(model, x, y, ndraws=10)
        self.assertTrue(
            os.path.isfile(
                "{}/{}_plot_with_data.png".format(self.result.outdir, self.result.label)
            )
        )
        self.result.posterior["log_likelihood"] = np.random.uniform(
            0, 1, len(self.result.posterior)
        )
        self.result.plot_with_data(model, x, y, ndraws=10, xlabel="a", ylabel="y")

    def test_plot_corner(self):
        self.result.injection_parameters = dict(x=0.8, y=1.1)
        self.result.plot_corner()
        self.result.plot_corner(parameters=["x", "y"])
        self.result.plot_corner(parameters=["x", "y"], truths=[1, 1])
        self.result.plot_corner(parameters=dict(x=1, y=1))
        self.result.plot_corner(truths=dict(x=1, y=1))
        self.result.plot_corner(truth=dict(x=1, y=1))
        self.result.plot_corner(truths=None)
        self.result.plot_corner(truths=False)
        self.result.plot_corner(truths=True)
        with self.assertRaises(ValueError):
            self.result.plot_corner(truths=dict(x=1, y=1), parameters=dict(x=1, y=1))
        with self.assertRaises(ValueError):
            self.result.plot_corner(truths=[1, 1], parameters=dict(x=1, y=1))
        with self.assertRaises(ValueError):
            self.result.plot_corner(parameters=["x", "y"], truths=dict(x=1, y=1))

    def test_plot_corner_with_injection_parameters(self):
        self.result.plot_corner()
        self.result.plot_corner(parameters=["x", "y"])
        self.result.plot_corner(parameters=["x", "y"], truths=[1, 1])
        self.result.plot_corner(parameters=dict(x=1, y=1))

    def test_plot_corner_with_priors(self):
        priors = bilby.core.prior.PriorDict()
        priors["x"] = bilby.core.prior.Uniform(-1, 1, "x")
        priors["y"] = bilby.core.prior.Uniform(-1, 1, "y")
        self.result.plot_corner(priors=priors)
        self.result.priors = priors
        self.result.plot_corner(priors=True)
        with self.assertRaises(ValueError):
            self.result.plot_corner(priors="test")

    def test_get_credible_levels(self):
        levels = self.result.get_all_injection_credible_levels()
        self.assertDictEqual(levels, dict(x=0.68, y=0.72))

    def test_get_credible_levels_raises_error_if_no_injection_parameters(self):
        self.result.injection_parameters = None
        with self.assertRaises(TypeError):
            self.result.get_all_injection_credible_levels()

    def test_kde(self):
        kde = self.result.kde
        import scipy.stats

        self.assertEqual(type(kde), scipy.stats.kde.gaussian_kde)
        self.assertEqual(kde.d, 2)

    def test_posterior_probability(self):
        sample = dict(x=0, y=0.1)
        self.assertTrue(
            isinstance(self.result.posterior_probability(sample), np.ndarray)
        )
        self.assertTrue(len(self.result.posterior_probability(sample)), 1)
        self.assertEqual(
            self.result.posterior_probability(sample)[0], self.result.kde([0, 0.1])
        )

    def test_multiple_posterior_probability(self):
        sample = [dict(x=0, y=0.1), dict(x=0.8, y=0)]
        self.assertTrue(
            isinstance(self.result.posterior_probability(sample), np.ndarray)
        )
        self.assertTrue(
            np.array_equal(
                self.result.posterior_probability(sample),
                self.result.kde([[0, 0.1], [0.8, 0]]),
            )
        )

    def test_to_arviz(self):
        with self.assertRaises(TypeError):
            self.result.to_arviz(prior=dict())

        Nprior = 100

        log_likelihood = np.random.rand(len(self.result.posterior))
        self.result.log_likelihood_evaluations = log_likelihood

        az = self.result.to_arviz(prior=Nprior)

        self.assertTrue("x" in az.posterior and "y" in az.posterior)
        for var in ["x", "y"]:
            self.assertTrue(np.array_equal(az.posterior[var].values.squeeze(),
                                           self.result.posterior[var].values))
            self.assertTrue(len(az.prior[var][0]) == Nprior)

        self.assertTrue(np.array_equal(az.log_likelihood["log_likelihood"].values.squeeze(),
                                       log_likelihood))

        self.assertTrue(
            az.posterior.attrs["inference_library"] == "bilby: {}".format(
                self.result.sampler
            )
        )
        self.assertTrue(
            az.posterior.attrs["inference_library_version"]
            == bilby.utils.get_version_information()
        )

        # add log likelihood to samples and extract from there
        del az
        self.result.posterior["log_likelihood"] = log_likelihood
        az = self.result.to_arviz()
        self.assertTrue(np.array_equal(az.log_likelihood["log_likelihood"].values.squeeze(),
                                       log_likelihood))


class TestResultListError(unittest.TestCase):
    def setUp(self):
        np.random.seed(7)
        bilby.utils.command_line_args.bilby_test_mode = False
        self.priors = bilby.prior.PriorDict(
            dict(
                x=bilby.prior.Uniform(0, 1, "x", latex_label="$x$", unit="s"),
                y=bilby.prior.Uniform(0, 1, "y", latex_label="$y$", unit="m"),
                c=1,
                d=2,
            )
        )

        # create two cpnest results
        self.nested_results = bilby.result.ResultList([])
        self.mcmc_results = bilby.result.ResultList([])
        self.expected_nested = []

        self.outdir = "outdir"
        self.label = "label"
        n = 100
        for i in range(2):
            result = bilby.core.result.Result(
                label=self.label + str(i),
                outdir=self.outdir,
                sampler="cpnest",
                search_parameter_keys=["x", "y"],
                fixed_parameter_keys=["c", "d"],
                priors=self.priors,
                sampler_kwargs=dict(test="test", func=lambda x: x, nlive=10),
                injection_parameters=dict(x=0.5, y=0.5),
                meta_data=dict(test="test"),
            )

            posterior = pd.DataFrame(
                dict(
                    x=np.random.normal(0, 1, n),
                    y=np.random.normal(0, 1, n),
                    log_likelihood=sorted(np.random.normal(0, 1, n)),
                )
            )
            result.posterior = posterior[-10:]  # use last 10 samples as posterior
            result.nested_samples = posterior
            result.log_evidence = 10
            result.log_evidence_err = 11
            result.log_bayes_factor = 12
            result.log_noise_evidence = 13
            self.nested_results.append(result)
            self.expected_nested.append(result)
        for i in range(2):
            result = bilby.core.result.Result(
                label=self.label + str(i) + "mcmc",
                outdir=self.outdir,
                sampler="emcee",
                search_parameter_keys=["x", "y"],
                fixed_parameter_keys=["c", "d"],
                priors=self.priors,
                sampler_kwargs=dict(test="test", func=lambda x: x, nlive=10),
                injection_parameters=dict(x=0.5, y=0.5),
                meta_data=dict(test="test"),
            )

            posterior = pd.DataFrame(
                dict(
                    x=np.random.normal(0, 1, n),
                    y=np.random.normal(0, 1, n),
                    log_likelihood=sorted(np.random.normal(0, 1, n)),
                )
            )
            result.posterior = posterior[-10:]
            result.log_evidence = 10
            result.log_evidence_err = 11
            result.log_bayes_factor = 12
            result.log_noise_evidence = 13
            self.mcmc_results.append(result)
        for res in self.nested_results:
            res.save_to_file()

    def tearDown(self):
        bilby.utils.command_line_args.bilby_test_mode = True
        try:
            shutil.rmtree(self.nested_results[0].outdir)
        except OSError:
            pass

        del self.nested_results
        del self.label
        del self.outdir
        del self.expected_nested
        del self.mcmc_results
        del self.priors

    def test_append_illegal_type(self):
        with self.assertRaises(TypeError):
            _ = bilby.core.result.ResultList(1)

    def test_append_from_string(self):
        self.nested_results.append(self.outdir + "/" + self.label + "1_result.json")
        self.assertEqual(3, len(self.nested_results))

    def test_append_result_type(self):
        self.nested_results.append(self.nested_results[1])
        self.expected_nested.append(self.expected_nested[1])
        self.assertListEqual(self.nested_results, self.nested_results)

    def test_combine_inconsistent_samplers(self):
        self.nested_results[0].sampler = "dynesty"
        with self.assertRaises(bilby.result.ResultListError):
            self.nested_results.combine()

    def test_combine_inconsistent_priors_length(self):
        self.nested_results[0].priors = bilby.prior.PriorDict(
            dict(
                x=bilby.prior.Uniform(0, 1, "x", latex_label="$x$", unit="s"),
                y=bilby.prior.Uniform(0, 1, "y", latex_label="$y$", unit="m"),
                c=1,
            )
        )
        with self.assertRaises(bilby.result.ResultListError):
            self.nested_results.combine()

    def test_combine_inconsistent_priors_types(self):
        self.nested_results[0].priors = bilby.prior.PriorDict(
            dict(
                x=bilby.prior.Uniform(0, 1, "x", latex_label="$x$", unit="s"),
                y=bilby.prior.Uniform(0, 1, "y", latex_label="$y$", unit="m"),
                c=1,
                d=bilby.core.prior.Cosine(),
            )
        )
        with self.assertRaises(bilby.result.ResultListError):
            self.nested_results.combine()

    def test_combine_inconsistent_search_parameters(self):
        self.nested_results[0].search_parameter_keys = ["y"]
        with self.assertRaises(bilby.result.ResultListError):
            self.nested_results.combine()

    def test_combine_inconsistent_data(self):
        self.nested_results[0].log_noise_evidence = -7
        with self.assertRaises(bilby.result.ResultListError):
            self.nested_results.combine()

    def test_combine_inconsistent_data_nan(self):
        self.nested_results[0].log_noise_evidence = np.nan
        self.nested_results[1].log_noise_evidence = np.nan
        self.nested_results.combine()

    def test_combine_inconsistent_sampling_data(self):
        result = bilby.core.result.Result(
            label=self.label,
            outdir=self.outdir,
            sampler="cpnest",
            search_parameter_keys=["x", "y"],
            fixed_parameter_keys=["c", "d"],
            priors=self.priors,
            sampler_kwargs=dict(test="test", func=lambda x: x, nlive=10),
            injection_parameters=dict(x=0.5, y=0.5),
            meta_data=dict(test="test"),
        )

        posterior = pd.DataFrame(
            dict(
                x=np.random.normal(0, 1, 100),
                y=np.random.normal(0, 1, 100),
                log_likelihood=sorted(np.random.normal(0, 1, 100)),
            )
        )
        result.posterior = posterior[-10:]  # use last 10 samples as posterior
        result.log_evidence = 10
        result.log_evidence_err = 11
        result.log_bayes_factor = 12
        result.log_noise_evidence = 13
        result._nested_samples = None
        self.nested_results.append(result)
        with self.assertRaises(bilby.result.ResultListError):
            self.nested_results.combine()


class TestMiscResults(unittest.TestCase):
    def test_sanity_check_labels(self):
        labels = ["a", "$a$", "a_1", "$a_1$"]
        labels_checked = bilby.core.result.sanity_check_labels(labels)
        self.assertEqual(labels_checked, ["a", "$a$", "a-1", "$a_1$"])


if __name__ == "__main__":
    unittest.main()
